import os

from lemoncheesecake.cli.command import Command
from lemoncheesecake.cli.utils import load_suites_from_project, add_project_cli_arg
from lemoncheesecake.exceptions import LemoncheesecakeException, ProjectNotFound, UserError, serialize_current_exception
from lemoncheesecake.filter import add_test_filter_cli_args, make_test_filter
from lemoncheesecake.project import load_project, PreparedProject, DEFAULT_REPORTING_BACKENDS
from lemoncheesecake.reporting.backend import get_reporting_backend_names as do_get_reporting_backend_names, \
    parse_reporting_backend_names_expression, get_reporting_backends_for_test_run
from lemoncheesecake.reporting.savingstrategy import make_report_saving_strategy, DEFAULT_REPORT_SAVING_STRATEGY


def get_nb_threads(cli_args, project):
    if cli_args.threads is not None:
        nb_threads = max(cli_args.threads, 1)
    elif "LCC_THREADS" in os.environ:
        try:
            nb_threads = max(int(os.environ["LCC_THREADS"]), 1)
        except ValueError:
            raise LemoncheesecakeException(
                "Invalid value '%s' for $LCC_THREADS environment variable (expect integer)" % os.environ["LCC_THREADS"]
            )
    else:
        nb_threads = 1

    if nb_threads > 1 and not project.threaded:
        raise LemoncheesecakeException("Project does not support multi-threading")

    return nb_threads


def get_report_saving_strategy(cli_args):
    saving_strategy_expression = cli_args.save_report or \
        os.environ.get("LCC_SAVE_REPORT") or DEFAULT_REPORT_SAVING_STRATEGY
    try:
        return make_report_saving_strategy(saving_strategy_expression)
    except ValueError as excp:
        raise LemoncheesecakeException(str(excp))


def get_reporting_backend_names(cli_args, project):
    if cli_args.reporting:
        try:
            return do_get_reporting_backend_names(
                project.default_reporting_backend_names, cli_args.reporting
            )
        except ValueError as e:
            raise LemoncheesecakeException("Invalid --reporting argument: %s" % e)
    elif "LCC_REPORTING" in os.environ:
        try:
            return do_get_reporting_backend_names(
                project.default_reporting_backend_names,
                parse_reporting_backend_names_expression(os.environ["LCC_REPORTING"])
            )
        except ValueError as e:
            raise LemoncheesecakeException("Invalid $LCC_REPORTING: %s" % e)
    else:
        return project.default_reporting_backend_names


def create_report_dir(cli_args, project):
    report_dir = cli_args.report_dir or os.environ.get("LCC_REPORT_DIR")
    if report_dir:
        try:
            os.mkdir(report_dir)
        except OSError as e:
            return LemoncheesecakeException("Cannot create report directory: %s" % e)
    else:
        try:
            report_dir = project.create_report_dir()
        except UserError as e:
            raise e
        except Exception:
            raise LemoncheesecakeException(
                "Got an unexpected exception while creating report directory:%s" % \
                    serialize_current_exception(show_stacktrace=True)
            )

    return report_dir


def run_suites_from_project(project, cli_args):
    # Create prepared_project
    prepared_project = PreparedProject.create(
        project, load_suites_from_project(project, make_test_filter(cli_args)), cli_args
    )

    # Get reporting backends
    reporting_backends = get_reporting_backends_for_test_run(
        project.reporting_backends, get_reporting_backend_names(cli_args, project)
    )

    # Get report save mode
    report_saving_strategy = get_report_saving_strategy(cli_args)

    # Create report dir
    report_dir = create_report_dir(cli_args, project)

    # Get number of threads
    nb_threads = get_nb_threads(cli_args, project)

    # Run tests
    report = prepared_project.run(
        reporting_backends, report_dir, report_saving_strategy,
        cli_args.force_disabled, cli_args.stop_on_failure, nb_threads
    )

    # Return exit code
    if cli_args.exit_error_on_failure:
        return 0 if report.is_successful() else 1
    else:
        return 0


class RunCommand(Command):
    def get_name(self):
        return "run"

    def get_description(self):
        return "Run the tests"

    def add_cli_args(self, cli_parser):
        try:
            project = load_project()
            default_reporting_backend_names = project.default_reporting_backend_names
        except ProjectNotFound:
            project = None
            default_reporting_backend_names = DEFAULT_REPORTING_BACKENDS

        add_test_filter_cli_args(cli_parser)

        project_group = cli_parser.add_argument_group("Project")
        add_project_cli_arg(project_group)

        test_execution_group = cli_parser.add_argument_group("Test execution")
        test_execution_group.add_argument(
            "--force-disabled", action="store_true",
            help="Force the run of disabled tests"
        )
        test_execution_group.add_argument(
            "--exit-error-on-failure", action="store_true",
            help="Exit with non-zero code if there is at least one non-passed test"
        )
        test_execution_group.add_argument(
            "--stop-on-failure", action="store_true",
            help="Stop tests execution on the first non-passed test"
        )
        test_execution_group.add_argument(
            "--threads", type=int, default=None,
            help="Number of threads used to run tests (default: $LCC_THREADS or 1)"
        )

        reporting_group = cli_parser.add_argument_group("Reporting")
        reporting_group.add_argument(
            "--report-dir", "-r", required=False,
            help="Directory where report data will be stored (default: $LCC_REPORT_DIR or 'report' depending on "
                 "the project configuration)"
        )
        reporting_group.add_argument(
            "--reporting", nargs="+", default=[],
            help="The list of reporting backends to use (default: %s)" % ", ".join(default_reporting_backend_names)
        )
        reporting_group.add_argument(
            "--save-report", required=False,
            help="At what frequency the reporting backends such as json or xml must save reporting data to disk. "
                 "(default: $LCC_SAVE_REPORT or at_each_failed_test, possible values are: "
                 "at_end_of_tests, at_each_suite, at_each_test, at_each_failed_test, at_each_log, every_${N}s)"
        )

        if project:
            cli_group = cli_parser.add_argument_group("Project custom arguments")
            project.add_cli_args(cli_group)

    def run_cmd(self, cli_args):
        return run_suites_from_project(load_project(cli_args.project), cli_args)



import os
import os.path as osp
import inspect

from typing import Sequence, Any, List

from lemoncheesecake.helpers.moduleimport import get_matching_files, get_py_files_from_dir, strip_py_ext, import_module
from lemoncheesecake.helpers.introspection import get_object_attributes
from lemoncheesecake.exceptions import LemoncheesecakeException, UserError, ModuleImportError, \
    SuiteLoadingError, serialize_current_exception
from lemoncheesecake.suite.core import Test, Suite, SUITE_HOOKS
from lemoncheesecake.suite.builder import _get_metadata_next_rank, build_description_from_name
from lemoncheesecake.testtree import BaseTreeNode
from lemoncheesecake.helpers.typecheck import check_type_string, check_type_dict, check_type


def _is_suite_class(obj):
    return inspect.isclass(obj) and hasattr(obj, "_lccmetadata") and obj._lccmetadata.is_suite


def _is_test_method(obj):
    return inspect.ismethod(obj) and hasattr(obj, "_lccmetadata") and obj._lccmetadata.is_test


def _is_test_function(obj):
    return inspect.isfunction(obj) and hasattr(obj, "_lccmetadata") and obj._lccmetadata.is_test


def _check_tags(tags):
    return isinstance(tags, (list, tuple)) and \
           all(isinstance(tag, str) for tag in tags)


def _check_properties(props):
    return isinstance(props, dict) and \
        all(isinstance(key, str) for key in props.keys()) and \
        all(isinstance(value, str) for value in props.values())


def _check_links(links):
    return isinstance(links, (list, tuple)) and \
        all(isinstance(url, str) for url, _ in links) and \
        all(name is None or isinstance(name, str) for _, name in links)


def _check_test_tree_node_types(node: BaseTreeNode) -> None:
    check_type_string("name", node.name)
    check_type_string("description", node.description)
    check_type("tags", node.tags, "List[str]", _check_tags, show_actual_type=False)
    check_type("properties", node.properties, "Dict[str, str]", _check_properties, show_actual_type=False)
    check_type("links", node.links, "List[Tuple[str, Optional[str]]]", _check_links, show_actual_type=False)


def _load_test(obj):
    md = obj._lccmetadata

    test = Test(md.name, md.description, obj)
    test.tags.extend(md.tags)
    test.properties.update(md.properties)
    test.links.extend(md.links)
    test.disabled = md.disabled
    test.hidden = md.condition and not md.condition(obj)
    test.rank = md.rank
    test.dependencies.extend(md.dependencies)

    try:
        _check_test_tree_node_types(test)
    except TypeError as excp:
        raise SuiteLoadingError("Invalid test metadata type for '%s': %s" % (test.name, excp))

    return test


def _load_parametrized_tests(obj):
    md = obj._lccmetadata
    test = _load_test(obj)

    for idx, parameters in enumerate(md.parametrized.parameters_source):
        check_type_dict("test parameters", parameters)
        parametrized_test_name, parametrized_test_description = md.parametrized.naming_scheme(
            test.name, test.description, parameters, idx+1
        )
        parametrized_test = test.pull_node()
        parametrized_test.name = parametrized_test_name
        parametrized_test.description = parametrized_test_description
        parametrized_test.parameters = parameters

        yield parametrized_test


def _load_tests(objs):
    for obj in objs:
        if obj._lccmetadata.parametrized:
            for test in _load_parametrized_tests(obj):
                if not test.hidden:
                    yield test
        else:
            test = _load_test(obj)
            if not test.hidden:
                yield test


def _get_test_symbols(obj, filter_func):
    return sorted(
        filter(
            filter_func, (attr for _, attr in get_object_attributes(obj))
        ),
        key=lambda sym: sym._lccmetadata.rank
    )


def _get_test_methods_from_class(obj):
    return _get_test_symbols(obj, _is_test_method)


def _get_sub_suites_from_class(obj):
    return _get_test_symbols(obj, _is_suite_class)


def _get_test_functions_from_module(mod):
    return _get_test_symbols(mod, _is_test_function)


def _get_suite_classes_from_module(mod):
    return _get_test_symbols(mod, _is_suite_class)


def _get_generated_tests(obj):
    return getattr(obj, "_lccgeneratedtests", [])


def _get_sub_dirs_from_dir(top_dir):
    paths = [os.path.join(top_dir, path) for path in os.listdir(top_dir)]
    return sorted(filter(osp.isdir, paths))


def _normalize_link(link):
    if type(link) is str:
        return link, None
    else:
        return link


def load_suite_from_class(class_: Any) -> Suite:
    """
    Load a suite from a class.
    """
    try:
        md = class_._lccmetadata
    except AttributeError:
        raise SuiteLoadingError("Class is not declared as a suite")

    try:
        suite_obj = class_()
    except UserError as e:
        raise e  # propagate UserError
    except Exception:
        raise LemoncheesecakeException("Got an unexpected error while instantiating suite class '%s':%s" % (
            class_.__name__, serialize_current_exception()
        ))

    suite = Suite(suite_obj, md.name, md.description)
    suite.tags.extend(md.tags)
    suite.properties.update(md.properties)
    suite.links.extend(md.links)
    suite.rank = md.rank
    suite.disabled = md.disabled
    suite.hidden = md.condition and not md.condition(suite_obj)

    try:
        _check_test_tree_node_types(suite)
    except TypeError as excp:
        raise SuiteLoadingError("Invalid suite metadata type for '%s': %s" % (suite.name, excp))

    for hook_name in SUITE_HOOKS:
        if hasattr(suite_obj, hook_name):
            suite.add_hook(hook_name, getattr(suite_obj, hook_name))

    for test in _load_tests(_get_test_methods_from_class(suite_obj)):
        suite.add_test(test)

    for test in _get_generated_tests(suite_obj):
        suite.add_test(test)

    for sub_suite in load_suites_from_classes(_get_sub_suites_from_class(suite_obj)):
        suite.add_suite(sub_suite)

    return suite


def load_suites_from_classes(classes: Sequence[Any]) -> List[Suite]:
    """
    Load a list of suites from a list of classes.
    """
    return list(
        filter(
            lambda suite: not suite.hidden, map(load_suite_from_class, classes)
        )
    )


def load_suite_from_module(mod: Any) -> Suite:
    """
    Load a suite from a module instance.
    """

    suite_info = getattr(mod, "SUITE", {})
    suite_condition = suite_info.get("visible_if")
    suite_name = suite_info.get("name", inspect.getmodulename(inspect.getfile(mod)))
    suite_description = suite_info.get("description", build_description_from_name(suite_name))

    suite = Suite(mod, suite_name, suite_description)
    suite.tags.extend(suite_info.get("tags", []))
    suite.properties.update(suite_info.get("properties", {}))
    suite.links.extend(map(_normalize_link, suite_info.get("links", [])))
    suite.rank = suite_info.get("rank", _get_metadata_next_rank())
    suite.hidden = suite_condition and not suite_condition(mod)

    try:
        _check_test_tree_node_types(suite)
    except TypeError as excp:
        raise SuiteLoadingError("Invalid suite metadata type for '%s': %s" % (suite.name, excp))

    for hook_name in SUITE_HOOKS:
        if hasattr(mod, hook_name):
            suite.add_hook(hook_name, getattr(mod, hook_name))

    for test in _load_tests(_get_test_functions_from_module(mod)):
        suite.add_test(test)

    for test in _get_generated_tests(mod):
        suite.add_test(test)

    for sub_suite in load_suites_from_classes(_get_suite_classes_from_module(mod)):
        suite.add_suite(sub_suite)

    return suite


def load_suite_from_file(filename: str) -> Suite:
    """
    Load a suite from a Python module indicated by a filename.

    Raise SuiteLoadingError if the file cannot be loaded as a suite.
    """
    try:
        mod = import_module(filename)
    except ModuleImportError as e:
        raise SuiteLoadingError(str(e))
    mod_name = strip_py_ext(osp.basename(filename))

    suite = load_suite_from_module(mod)

    # in case of module containing no tests, a single sub-suite whose name is the same a as module
    # then this sub-suite will be actually considered as the suite itself
    if not hasattr(mod, "SUITE") and \
        len(suite.get_tests()) == 0 and len(suite.get_suites()) == 1 and suite.get_suites()[0].name == mod_name:
        suite = suite.get_suites()[0]
        suite.parent_suite = None

    return suite


def load_suites_from_files(patterns: Sequence[str], excluding: Sequence[str] = ()) -> List[Suite]:
    """
    Load a list of suites from a list of files.

    :param patterns: a mandatory list (a simple string can also be used instead of a single element list)
      of files to import; the wildcard '*' character can be used
    :param excluding: an optional list (a simple string can also be used instead of a single element list)
      of elements to exclude from the expanded list of files to import

    Example::

        load_suites_from_files("test_*.py")
    """
    return list(
        filter(
            lambda suite: not suite.hidden and not suite.is_empty(),
            map(load_suite_from_file, get_matching_files(patterns, excluding))
        )
    )


def load_suites_from_directory(dir: str, recursive: bool = True) -> List[Suite]:
    """
    Load a list of suites from a directory.

    If the recursive argument is set to True, sub suites will be searched in a directory named
    from the suite module: if the suite module is "foo.py" then the sub suites directory must be "foo".

    Raise SuiteLoadingError if one or more suite could not be loaded.
    """
    if not osp.exists(dir):
        raise SuiteLoadingError("Directory '%s' does not exist" % dir)

    suites = {}

    for filename in get_py_files_from_dir(dir):
        suite = load_suite_from_file(filename)
        if not suite.hidden:
            suites[filename] = suite

    if recursive:
        for dirname in _get_sub_dirs_from_dir(dir):
            suite = suites.get(dirname + ".py")
            if not suite:
                suite_name = osp.basename(dirname)
                suite = Suite(None, suite_name, build_description_from_name(suite_name))
                suites[suite.name] = suite
            for sub_suite in load_suites_from_directory(dirname, recursive=True):
                suite.add_suite(sub_suite)

    return \
        sorted(
            sorted(
                filter(lambda s: not s.is_empty(), suites.values()),
                key=lambda s: s.name
            ),
            key=lambda s: s.rank
        )



import os.path as osp


def get_resource_path(relpath):
    if isinstance(relpath, (list, tuple)):
        relpath = osp.join(*relpath)

    return osp.join(osp.dirname(__file__), osp.pardir, "resources", relpath)



'''
Created on Apr 3, 2017

@author: nicolas
'''

from typing import List, Any, Union

from lemoncheesecake.helpers.text import jsonify
from lemoncheesecake.matching.matcher import Matcher, MatchResult, MatcherDescriptionTransformer
from lemoncheesecake.matching.matchers.value import is_

_TYPE_NAMES = {
    int: "integer",
    float: "float",
    str: "string",
    dict: "collection",
    list: "array", tuple: "array"
}


class IsValueOfType(Matcher):
    def __init__(self, types: List[Any], type_name: str, value_matcher: Matcher) -> None:
        self.types = types
        self.type_name = type_name
        self.value_matcher = value_matcher

    def build_description(self, transformation):
        ret = transformation("to be %s" % self.type_name)
        if self.value_matcher:
            ret += " that %s" % self.value_matcher.build_description(MatcherDescriptionTransformer(conjugate=True))

        return ret

    @staticmethod
    def _get_value_type_name(value):
        try:
            return _TYPE_NAMES[type(value)]
        except KeyError:
            return str(type(value))

    def matches(self, actual):
        if type(actual) in self.types:
            if self.value_matcher:
                return self.value_matcher.matches(actual)
            else:
                return MatchResult.success("got %s" % jsonify(actual))
        else:
            return MatchResult.failure("got %s (%s)" % (jsonify(actual), self._get_value_type_name(actual)))


def _is_type(types, type_name, value_matcher):
    return IsValueOfType(
        types, type_name, is_(value_matcher) if value_matcher is not None else None
    )


def is_integer(expected: Union[int, Matcher] = None) -> Matcher:
    """
    Test if value is an integer.
    """
    return _is_type([int], "an integer", expected)


def is_float(expected: Union[float, Matcher] = None) -> Matcher:
    """
    Test if value is a float.
    """
    return _is_type([float], "a float", expected)


def is_bool(expected: Union[bool, Matcher] = None) -> Matcher:
    """
    Test if value is a boolean.
    """
    return _is_type([bool], "a boolean", expected)


def is_str(expected: Union[str, Matcher] = None) -> Matcher:
    """
    Test if value is a string.
    """
    return _is_type([str], "a string", expected)


def is_dict(expected: Union[dict, Matcher] = None) -> Matcher:
    """
    Test if value is a dict (key/value collection).
    """
    return _is_type([dict], "a collection", expected)


def is_list(expected: Union[list, tuple, Matcher] = None) -> Matcher:
    """
    Test if value is a list.
    """
    return _is_type([list, tuple], "a list", expected)



'''
Created on Mar 27, 2017

@author: nicolas
'''

from typing import Any, Mapping, Sequence, Optional

from lemoncheesecake.session import log_check
from lemoncheesecake.exceptions import AbortTest
from lemoncheesecake.matching.matcher import Matcher, MatchResult, MatcherDescriptionTransformer
from lemoncheesecake.matching.matchers.dict_ import HasEntry, KeyPathMatcher


class _HasEntry(HasEntry):
    def build_description(self, _):
        ret = self.key_matcher.build_description()
        if self.value_matcher:
            ret += " " + self.value_matcher.build_description(MatcherDescriptionTransformer())
        return ret


def _build_has_entry_matchers_from_arg(arg, path=()):
    if isinstance(arg, (list, tuple)):
        for idx, value in enumerate(arg):
            yield from _build_has_entry_matchers_from_arg(value, path + (idx,))

    elif isinstance(arg, dict):
        for key, value in arg.items():
            yield from _build_has_entry_matchers_from_arg(value, path + (key,))

    elif isinstance(arg, Matcher):
        yield _HasEntry(KeyPathMatcher(path), arg)

    else:
        raise ValueError("argument %r must be either an instance of list, tuple, dict or Matcher" % arg)


def _normalize_key_path(path):
    if isinstance(path, tuple):
        return path
    elif isinstance(path, list):
        return tuple(path)
    else:
        return (path,)


def _build_has_entry_matchers_from_args(args, base_key=()):
    base_key = _normalize_key_path(base_key)

    if len(args) == 1:
        assert isinstance(args[0], (list, tuple, dict))
        yield from _build_has_entry_matchers_from_arg(args[0], base_key)

    elif len(args) % 2 == 0:
        i = 0
        while i < len(args):
            key, value_matcher = args[i], args[i + 1]
            yield from _build_has_entry_matchers_from_arg(value_matcher, base_key + _normalize_key_path(key))
            i += 2

    else:
        raise ValueError(
            "*args must be a list of one element (being a list/tuple/dict) or "
            "a even number list (a list of key/matcher pairs)"
        )


def _format_result_details(details):
    if details is None:
        return None

    if not isinstance(details, str):
        details = str(details)

    return details[0].upper() + details[1:]


def _log_match_result(hint, matcher, result, quiet=False):
    if hint is not None:
        description = "Expect %s %s" % (hint, matcher.build_description(MatcherDescriptionTransformer()))
    else:
        description = "Expect %s" % matcher.build_description(MatcherDescriptionTransformer())

    return log_check(
        description, result.is_successful, _format_result_details(result.description) if not quiet else None
    )


def check_that(hint: Optional[str], actual: Any, matcher: Matcher, quiet: bool = False) -> MatchResult:
    """Check that actual matches given matcher.

    A check log is added to the report.

    If ``quiet`` is set to ``True``, the check details won't appear in the check log.
    """
    assert isinstance(matcher, Matcher)

    result = matcher.matches(actual)
    _log_match_result(hint, matcher, result, quiet=quiet)
    return result


def check_that_in(actual: Mapping, *expected, **kwargs) -> Sequence[MatchResult]:
    """
    Equivalent of :py:func:`check_that` over items of a dict.

    Example of usage::

        check_that_in(
            {"foo": 1, "bar": 2},
            "foo", equal_to(1),
            "bar", equal_to(2)
        )

    The key can also be a ``tuple`` when checking for a nested item::

        check_that_in(
            {"foo": {"bar": 2}},
            ("foo", "bar"), equal_to(2)
        )

    The function can take a ``base_key`` keyword-arg to pass repeating nested-keys as a ``tuple``.

    If an extra ``quiet`` keyword-arg is set to ``True``, the check details won't appear in the check log.

    The function returns a list of :py:class:`MatchResult <lemoncheesecake.matching.matcher.MatchResult>`.

    .. versionadded:: 1.11.0

    Nested data can also be checked this way::

        check_that_in(
            {"foo": {"bar": 2}},
            {"foo": {"bar": equal_to(2)}}
        )

    It also means that instead of calling ``check_that_in`` with a list of key/matcher pairs, it can now
    be called with a data structure as the "expected" argument.
    """
    quiet = kwargs.get("quiet", False)
    base_key = kwargs.get("base_key", ())

    return [
        check_that(None, actual, matcher, quiet=quiet)
        for matcher in _build_has_entry_matchers_from_args(expected, base_key)
    ]


def require_that_in(actual: Mapping, *expected, **kwargs) -> Sequence[MatchResult]:
    """
    Does the same thing as :py:func:`check_that_in` except it performs a
    :py:func:`require_that` on each key-matcher pair.
    """
    quiet = kwargs.get("quiet", False)
    base_key = kwargs.get("base_key", ())

    return [
        require_that(None, actual, matcher, quiet=quiet)
        for matcher in _build_has_entry_matchers_from_args(expected, base_key)
    ]


def assert_that_in(actual: Mapping, *expected, **kwargs) -> Sequence[MatchResult]:
    """
    Does the same thing as :py:func:`check_that_in` except it performs a
    :py:func:`assert_that` on each key-matcher pair.
    """
    quiet = kwargs.get("quiet", False)
    base_key = kwargs.get("base_key", ())

    return [
        assert_that(None, actual, matcher, quiet=quiet)
        for matcher in _build_has_entry_matchers_from_args(expected, base_key)
    ]


def require_that(hint: Optional[str], actual: Any, matcher: Matcher, quiet: bool = False) -> MatchResult:
    """Require that actual matches given matcher.

    A check log is added to the report. An AbortTest exception is raised if the check does not succeed.

    If ``quiet`` is set to ``True``, the check details won't appear in the check log.
    """
    assert isinstance(matcher, Matcher)

    result = matcher.matches(actual)
    _log_match_result(hint, matcher, result, quiet=quiet)
    if result:
        return result
    else:
        raise AbortTest("previous requirement was not fulfilled")


def assert_that(hint: Optional[str], actual: Any, matcher: Matcher, quiet: bool = False) -> MatchResult:
    """Assert that actual matches given matcher.

    If assertion fail, a check log is added to the report and an AbortTest exception is raised.

    If ``quiet`` is set to ``True``, the check details won't appear in the check log.
    """
    assert isinstance(matcher, Matcher)

    result = matcher.matches(actual)
    if result:
        return result
    else:
        _log_match_result(hint, matcher, result, quiet=quiet)
        raise AbortTest("assertion error")



'''
Created on Mar 27, 2016

@author: nicolas
'''

import time

try:
    from lxml import etree as ET
    from lxml.builder import E
    LXML_IS_AVAILABLE = True
except ImportError:
    LXML_IS_AVAILABLE = False

import lemoncheesecake
from lemoncheesecake.reporting.backend import FileReportBackend, ReportUnserializerMixin
from lemoncheesecake.reporting.report import (
    Report, Log, Check, Attachment, Url, Step, Result, TestResult, SuiteResult,
    format_time_as_iso8601, parse_iso8601_time
)
from lemoncheesecake.exceptions import ReportLoadingError

DEFAULT_INDENT_LEVEL = 4


# borrowed from http://stackoverflow.com/a/1239193
def indent_xml(elem, level=0, indent_level=DEFAULT_INDENT_LEVEL):
    i = "\n" + level * (" " * indent_level)
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + (" " * indent_level)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent_xml(elem, level+1, indent_level)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


def make_xml_node(name, *args):
    node = E(name)
    i = 0
    while i < len(args):
        attr_name, attr_value = args[i], args[i+1]
        node.attrib[attr_name] = attr_value
        i += 2
    return node


def make_xml_child(parent, name, *args):
    child = make_xml_node(name, *args)
    parent.append(child)
    return child


def _serialize_time(value):
    return format_time_as_iso8601(value)


def _serialize_bool(value):
    return "true" if value else "false"


def _serialize_steps(steps, xml_result):
    for step in steps:
        xml_step = make_xml_child(
            xml_result, "step",
            "description", step.description,
            "start-time", _serialize_time(step.start_time)
        )
        if step.end_time is not None:
            xml_step.attrib["end-time"] = _serialize_time(step.end_time)
        for log in step.get_logs():
            if isinstance(log, Log):
                xml_log = make_xml_child(
                    xml_step, "log",
                    "level", log.level,
                    "time", _serialize_time(log.time)
                )
                xml_log.text = log.message
            elif isinstance(log, Attachment):
                xml_attachment = make_xml_child(
                    xml_step, "attachment",
                    "description", log.description,
                    "as-image", _serialize_bool(log.as_image),
                    "time", _serialize_time(log.time)
                )
                xml_attachment.text = log.filename
            elif isinstance(log, Url):
                xml_url = make_xml_child(
                    xml_step, "url",
                    "description", log.description,
                    "time", _serialize_time(log.time)
                )
                xml_url.text = log.url
            else:  # TestCheck
                xml_check = make_xml_child(
                    xml_step, "check",
                    "description", log.description,
                    "is-successful", _serialize_bool(log.is_successful),
                    "time", _serialize_time(log.time)
                )
                xml_check.text = log.details


def _serialize_result(result, xml_result):
    if result.status:
        xml_result.attrib["status"] = result.status
    if result.status_details:
        xml_result.attrib["status-details"] = result.status_details
    xml_result.attrib["start-time"] = _serialize_time(result.start_time)
    if result.end_time is not None:
        xml_result.attrib["end-time"] = _serialize_time(result.end_time)
    _serialize_steps(result.get_steps(), xml_result)


def _serialize_node_metadata(obj, xml_node):
    xml_node.attrib["name"] = obj.name
    xml_node.attrib["description"] = obj.description

    for tag in obj.tags:
        xml_tag = make_xml_child(xml_node, "tag")
        xml_tag.text = tag
    for name, value in obj.properties.items():
        xml_property = make_xml_child(xml_node, "property", "name", name)
        xml_property.text = value
    for link in obj.links:
        xml_link = make_xml_child(xml_node, "link")
        if link[1]:
            xml_link.attrib["name"] = link[1]
        xml_link.text = link[0]


def _serialize_test_result(test):
    xml_test = make_xml_node("test")
    _serialize_node_metadata(test, xml_test)
    _serialize_result(test, xml_test)
    return xml_test


def _serialize_suite_result(suite):
    xml_suite = make_xml_node("suite")

    _serialize_node_metadata(suite, xml_suite)

    xml_suite.attrib["start-time"] = _serialize_time(suite.start_time)
    if suite.end_time is not None:
        xml_suite.attrib["end-time"] = _serialize_time(suite.end_time)

    # before suite
    if suite.suite_setup:
        _serialize_result(suite.suite_setup, make_xml_child(xml_suite, "suite-setup"))

    # tests
    xml_suite.extend(map(_serialize_test_result, suite.get_tests()))

    # sub suites
    xml_suite.extend(map(_serialize_suite_result, suite.get_suites()))

    # after suite
    if suite.suite_teardown:
        _serialize_result(suite.suite_teardown, make_xml_child(xml_suite, "suite-teardown"))

    return xml_suite


def serialize_report_as_xml_tree(report):
    xml_report = E("lemoncheesecake-report")
    xml_report.attrib["lemoncheesecake-version"] = lemoncheesecake.__version__
    xml_report.attrib["report-version"] = "1.1"
    xml_report.attrib["start-time"] = _serialize_time(report.start_time)
    if report.end_time is not None:
        xml_report.attrib["end-time"] = _serialize_time(report.end_time)
    xml_report.attrib["generation-time"] = _serialize_time(time.time())
    xml_report.attrib["nb-threads"] = str(report.nb_threads)

    xml_title = make_xml_child(xml_report, "title")
    xml_title.text = report.title
    for name, value in report.info:
        xml_info = make_xml_child(xml_report, "info", "name", name)
        xml_info.text = value

    if report.test_session_setup:
        _serialize_result(report.test_session_setup, make_xml_child(xml_report, "test-session-setup"))

    xml_report.extend(map(_serialize_suite_result, report.get_suites()))

    if report.test_session_teardown:
        _serialize_result(report.test_session_teardown, make_xml_child(xml_report, "test-session-teardown"))

    return xml_report


def serialize_report_as_string(report, indent_level=DEFAULT_INDENT_LEVEL):
    xml_report = serialize_report_as_xml_tree(report)
    indent_xml(xml_report, indent_level=indent_level)

    return ET.tostring(xml_report, pretty_print=True, encoding="unicode")


def save_report_into_file(report, filename, indent_level=DEFAULT_INDENT_LEVEL):
    content = serialize_report_as_string(report, indent_level)
    with open(filename, "w") as fh:
        fh.write(content)


def _unserialize_time(value):
    return parse_iso8601_time(value)


def _unserialize_bool(value):
    if value == "true":
        return True
    elif value == "false":
        return False
    else:
        raise ValueError("Invalid boolean representation: '%s'" % value)


def _unserialize_step(xml_step):
    step = Step(xml_step.attrib["description"])
    step.start_time = _unserialize_time(xml_step.attrib["start-time"])
    step.end_time = _unserialize_time(xml_step.attrib["end-time"]) if "end-time" in xml_step.attrib else None
    for xml_log in xml_step:
        if xml_log.tag == "log":
            step_log = Log(
                xml_log.attrib["level"], xml_log.text, _unserialize_time(xml_log.attrib["time"])
            )
        elif xml_log.tag == "attachment":
            step_log = Attachment(
                xml_log.attrib["description"], xml_log.text,
                _unserialize_bool(xml_log.attrib["as-image"]),
                _unserialize_time(xml_log.attrib["time"])
            )
        elif xml_log.tag == "url":
            step_log = Url(
                xml_log.attrib["description"], xml_log.text, _unserialize_time(xml_log.attrib["time"])
            )
        elif xml_log.tag == "check":
            step_log = Check(
                xml_log.attrib["description"], _unserialize_bool(xml_log.attrib["is-successful"]),
                xml_log.text, _unserialize_time(xml_log.attrib["time"])
            )
        else:
            raise ValueError("Unknown tag '%s' for step" % xml_log.tag)
        step.add_log(step_log)
    return step


def _unserialize_result(xml_result, result):
    result.status = xml_result.attrib.get("status", None)
    # status_details for non-test results has been introduced in report version 1.1:
    result.status_details = xml_result.attrib.get("status-details", None)
    result.start_time = _unserialize_time(xml_result.attrib["start-time"])
    result.end_time = _unserialize_time(xml_result.attrib["end-time"]) if "end-time" in xml_result.attrib else None
    for xml_step in xml_result.xpath("step"):
        result.add_step(_unserialize_step(xml_step))


def _unserialize_node_metadata(xml_node, node):
    node.tags = [n.text for n in xml_node.xpath("tag")]
    node.properties = {n.attrib["name"]: n.text for n in xml_node.xpath("property")}
    node.links = [(n.text, n.attrib.get("name", None)) for n in xml_node.xpath("link")]


def _unserialize_test_result(xml_result):
    test = TestResult(xml_result.attrib["name"], xml_result.attrib["description"])
    _unserialize_result(xml_result, test)
    _unserialize_node_metadata(xml_result, test)
    return test


def _unserialize_suite_result(xml_suite):
    suite = SuiteResult(xml_suite.attrib["name"], xml_suite.attrib["description"])
    suite.start_time = _unserialize_time(xml_suite.attrib["start-time"])
    suite.end_time = _unserialize_time(xml_suite.attrib["end-time"]) if "end-time" in xml_suite.attrib else None
    _unserialize_node_metadata(xml_suite, suite)

    xml_setup = xml_suite.xpath("suite-setup")
    xml_setup = xml_setup[0] if len(xml_setup) > 0 else None
    if xml_setup is not None:
        suite.suite_setup = Result()
        _unserialize_result(xml_setup, suite.suite_setup)

    for xml_test in xml_suite.xpath("test"):
        suite.add_test(_unserialize_test_result(xml_test))

    xml_teardown = xml_suite.xpath("suite-teardown")
    xml_teardown = xml_teardown[0] if len(xml_teardown) > 0 else None
    if xml_teardown is not None:
        suite.suite_teardown = Result()
        _unserialize_result(xml_teardown, suite.suite_teardown)

    for xml_suite in xml_suite.xpath("suite"):
        suite.add_suite(_unserialize_suite_result(xml_suite))

    return suite


def _unserialize_report(xml_report):
    report = Report()

    report.start_time = _unserialize_time(xml_report.attrib["start-time"])
    report.end_time = _unserialize_time(xml_report.attrib["end-time"]) if "end-time" in xml_report.attrib else None
    report.saving_time = _unserialize_time(xml_report.attrib["generation-time"]) if "generation-time" in xml_report.attrib else None
    report.nb_threads = int(xml_report.attrib["nb-threads"])
    report.title = xml_report.xpath("title")[0].text
    report.info = [(node.attrib["name"], node.text) for node in xml_report.xpath("info")]

    xml_setup = xml_report.xpath("test-session-setup")
    xml_setup = xml_setup[0] if len(xml_setup) else None
    if xml_setup is not None:
        report.test_session_setup = Result()
        _unserialize_result(xml_setup, report.test_session_setup)

    for xml_suite in xml_report.xpath("suite"):
        report.add_suite(_unserialize_suite_result(xml_suite))

    xml_teardown = xml_report.xpath("test-session-teardown")
    xml_teardown = xml_teardown[0] if len(xml_teardown) else None
    if xml_teardown is not None:
        report.test_session_teardown = Result()
        _unserialize_result(xml_teardown, report.test_session_teardown)

    return report


def load_report_from_file(filename):
    try:
        with open(filename, "r") as fh:
            xml = ET.parse(fh)
    except ET.LxmlError as e:
        raise ReportLoadingError(str(e))
    except IOError as e:
        raise e  # re-raise as-is
    try:
        root = xml.getroot().xpath("/lemoncheesecake-report")[0]
    except IndexError:
        raise ReportLoadingError("Cannot find lemoncheesecake-report element in XML")

    report_version = float(root.attrib["report-version"])
    if report_version >= 2.0:
        raise ReportLoadingError("Incompatible report version: got %s while 1.x is supported" % report_version)

    return _unserialize_report(root)


class XmlBackend(FileReportBackend, ReportUnserializerMixin):
    def __init__(self):
        self.indent_level = DEFAULT_INDENT_LEVEL

    def get_name(self):
        return "xml"

    def is_available(self):
        return LXML_IS_AVAILABLE

    def get_report_filename(self):
        return "report.xml"

    def save_report(self, filename, report):
        save_report_into_file(report, filename, self.indent_level)

    def load_report(self, path):
        report = load_report_from_file(path)
        report.bind(self, path)
        return report



'''
Created on Feb 14, 2017

@author: nicolas
'''

from lemoncheesecake.helpers.console import bold
from lemoncheesecake.helpers.text import ensure_single_line_text
from lemoncheesecake.cli.command import Command
from lemoncheesecake.cli.utils import load_suites_from_project
from lemoncheesecake.filter import add_test_filter_cli_args, make_test_filter
from lemoncheesecake.project import load_project
from lemoncheesecake.reporting.console import serialize_metadata


class TestTreeRenderer:
    def __init__(self, show_description=False, indent=4):
        self.show_description = show_description
        self.indent = indent

    def get_padding(self, depth):
        return " " * (depth * self.indent)

    def get_test_label(self, test):
        if self.show_description:
            return ensure_single_line_text(test.description)
        else:
            return test.path

    def get_suite_label(self, suite):
        if self.show_description:
            return ensure_single_line_text(suite.description)
        else:
            return suite.path

    def show_test(self, test, suite):
        md = serialize_metadata(test)
        padding = self.get_padding(suite.hierarchy_depth + 1)
        test_label = self.get_test_label(test)
        print("%s- %s%s" % (padding, test_label, " (%s)" % md if md else ""))

    def show_suite(self, suite):
        md = serialize_metadata(suite)
        padding = self.get_padding(suite.hierarchy_depth)
        suite_label = self.get_suite_label(suite)
        print("%s* %s%s" % (padding, bold(suite_label), " (%s)" % md if md else ""))

        for test in suite.get_tests():
            self.show_test(test, suite)

        for sub_suite in suite.get_suites():
            self.show_suite(sub_suite)

    def show_suites(self, suites):
        for suite in suites:
            self.show_suite(suite)


class ShowCommand(Command):
    def get_name(self):
        return "show"

    def get_description(self):
        return "Show the test tree"

    def add_cli_args(self, cli_parser):
        add_test_filter_cli_args(cli_parser)

        group = cli_parser.add_argument_group("Display")
        group.add_argument(
            "--show-description", "-d", action="store_true",
            help="Show suite and test descriptions instead of paths"
        )

    def run_cmd(self, cli_args):
        project = load_project()
        suites = load_suites_from_project(project, make_test_filter(cli_args))

        renderer = TestTreeRenderer(show_description=cli_args.show_description)
        renderer.show_suites(suites)
        
        return 0



import itertools
from multiprocessing.dummy import Pool, Queue

from lemoncheesecake.exceptions import LemoncheesecakeException, TaskFailure, serialize_current_exception

DEBUG = False
_KEYBOARD_INTERRUPT_ERROR_MESSAGE = "tests have been interrupted by the user"


def _debug(msg):
    if DEBUG:
        print(msg)


class TaskResultSuccess:
    pass


class TaskResultSkipped:
    def __init__(self, reason):
        self.reason = reason


class TaskResultFailure:
    def __init__(self, reason):
        self.reason = reason


class TaskResultException:
    def __init__(self, stacktrace):
        self.stacktrace = stacktrace


class BaseTask:
    def __init__(self):
        self.result = None

    def get_all_dependencies(self):
        return self.get_on_completion_dependencies() + self.get_on_success_dependencies()

    def get_on_success_dependencies(self):
        return []

    def get_on_completion_dependencies(self):
        return []

    def run(self, context):
        pass

    def skip(self, context, reason):
        pass


class TaskContext:
    def __init__(self):
        self._tasks_aborted = False

    def enable_task_abort(self):
        self._tasks_aborted = True

    def is_task_to_be_skipped(self, task):
        if self._tasks_aborted:
            return "tests have been manually stopped"
        else:
            return None


def pop_runnable_tasks(remaining_tasks, completed_tasks, nb_tasks):
    runnable_tasks = list(
        itertools.islice(
            filter(lambda t: set(t.get_all_dependencies()).issubset(completed_tasks), remaining_tasks),
            0, nb_tasks
        )
    )

    for task in runnable_tasks:
        remaining_tasks.remove(task)
        _debug("pop runnable task %s" % task)
        yield task


def run_task(task, context, completed_task_queue):
    _debug("run task %s" % task)
    try:
        task.run(context)
    except TaskFailure as excp:
        task.result = TaskResultFailure(str(excp))
    except Exception:
        task.result = TaskResultException(serialize_current_exception())
    else:
        task.result = TaskResultSuccess()

    completed_task_queue.put(task)


def handle_task(task, context, completed_task_queue):
    _debug("handle task %s" % task)

    # skip task on dependency failure if any
    for dep_task in task.get_on_success_dependencies():
        if not isinstance(dep_task.result, TaskResultSuccess):
            if isinstance(dep_task.result, (TaskResultFailure, TaskResultSkipped)):
                reason = dep_task.result.reason
            else:
                reason = None
            skip_task(task, context, completed_task_queue, reason)
            return

    # skip task on external trigger if any
    skip_reason = context.is_task_to_be_skipped(task)
    if skip_reason:
        skip_task(task, context, completed_task_queue, reason=skip_reason)
        return

    # run task when all conditions are met
    run_task(task, context, completed_task_queue)


def schedule_tasks_to_be_run(tasks, context, pool, completed_tasks_queue):
    for task in tasks:
        pool.apply_async(handle_task, args=(task, context, completed_tasks_queue))


def skip_task(task, context, completed_task_queue, reason=""):
    _debug("skip task %s" % task)
    try:
        task.skip(context, reason)
    except Exception:
        task.result = TaskResultException(serialize_current_exception())
    else:
        task.result = TaskResultSkipped(reason)

    completed_task_queue.put(task)


def skip_all_tasks(tasks, remaining_tasks, completed_tasks, context, pool, completed_tasks_queue, reason):
    # schedule all tasks to be skipped...
    for task in remaining_tasks:
        pool.apply_async(skip_task, args=(task, context, completed_tasks_queue, reason))

    # ... and wait for their completion
    while len(completed_tasks) != len(tasks):
        completed_task = completed_tasks_queue.get()
        completed_tasks.add(completed_task)


def run_tasks(tasks, context, nb_threads=1):
    for task in tasks:
        check_task_dependencies(task)

    remaining_tasks = list(tasks)
    completed_tasks = set()

    pool = Pool(nb_threads)
    completed_tasks_queue = Queue()

    try:
        schedule_tasks_to_be_run(
            pop_runnable_tasks(remaining_tasks, completed_tasks, nb_threads),
            context, pool, completed_tasks_queue
        )

        while len(completed_tasks) != len(tasks):
            # wait for one task to complete
            completed_task = completed_tasks_queue.get()
            completed_tasks.add(completed_task)

            # schedule tasks to be run waiting for task success or simple completion
            tasks_to_be_run = pop_runnable_tasks(remaining_tasks, completed_tasks, nb_threads)
            schedule_tasks_to_be_run(tasks_to_be_run, context, pool, completed_tasks_queue)

    except KeyboardInterrupt:
        context.enable_task_abort()
        skip_all_tasks(
            tasks, remaining_tasks, completed_tasks, context, pool, completed_tasks_queue,
            _KEYBOARD_INTERRUPT_ERROR_MESSAGE
        )

    finally:
        pool.close()

    exceptions = [task.result.stacktrace for task in tasks if isinstance(task.result, TaskResultException)]
    if exceptions:
        raise LemoncheesecakeException(
            "Error(s) while running tasks, got exceptions:\n%s" % "\n".join(exceptions)
        )


def check_task_dependencies(task, task_path=()):
    for task_in_path in task_path:
        if task_in_path in task.get_all_dependencies():
            raise AssertionError("Task %s has a circular dependency on task %s" % (task, task_in_path))

    for dependency in task.get_all_dependencies():
        check_task_dependencies(dependency, task_path=(task,) + task_path)



'''
Created on Mar 26, 2016

@author: nicolas
'''

from __future__ import annotations

import time
from functools import reduce
from typing import Union, List, Iterator, Iterable, Optional, Callable
from datetime import datetime, timezone

from lemoncheesecake.helpers.time import humanize_duration
from lemoncheesecake.testtree import BaseTest, BaseSuite, flatten_tests, flatten_suites, find_test, find_suite, \
    normalize_node_hierarchy, TreeNodeHierarchy
from lemoncheesecake.reporting.backend import ReportSerializerMixin


def format_time_as_iso8601(ts: float) -> str:
    """
    Serialize time as ISO8601, such as: "2019-05-04T22:57:08.399Z"
    """
    ts = round(ts, 3)  # round the timestamp to milliseconds
    dt = datetime.utcfromtimestamp(ts)
    return dt.isoformat(timespec='milliseconds') + "Z"


def parse_iso8601_time(s: str) -> float:
    """
    Parse time as generated by format_time_as_iso8601
    """
    dt = datetime.fromisoformat(s.rstrip("Z"))
    return dt.replace(tzinfo=timezone.utc).timestamp()


def _get_duration(start_time: Optional[float], end_time: Optional[float]) -> Optional[float]:
    if start_time is not None and end_time is not None:
        return end_time - start_time
    else:
        return None


class StepLog:
    """
    Base class for logs contained in a Step instance.
    """
    def __init__(self, ts: float) -> None:
        #: Log time.
        self.time: float = ts
        #: Parent step.
        self.parent_step: Optional[Step] = None


class Log(StepLog):
    """
    The log resulting of log_info/log_debug/log_warning/log_error functions.
    Inherits :py:class:`StepLog <lemoncheesecake.reporting.StepLog>`.
    """
    LEVEL_DEBUG = "debug"
    LEVEL_INFO = "info"
    LEVEL_WARN = "warn"
    LEVEL_ERROR = "error"

    def __init__(self, level: str, message: str, ts: float) -> None:
        super().__init__(ts)
        #: Log level.
        self.level: str = level
        #: Log message.
        self.message: str = message


class Check(StepLog):
    """
    The log resulting of a check_that/require_that/assert_that functions.
    Inherits :py:class:`StepLog <lemoncheesecake.reporting.StepLog>`.
    """
    def __init__(self, description: str, is_successful: bool, details: Optional[str], ts: float) -> None:
        super().__init__(ts)
        #: Check description.
        self.description: str = description
        #: Whether the check is successful or not (boolean).
        self.is_successful: bool = is_successful
        #: Optional check details.
        self.details: str = details


class Attachment(StepLog):
    """
    The log resulting of save/prepare_*attachment* functions.
    Inherits :py:class:`StepLog <lemoncheesecake.reporting.StepLog>`.
    """
    def __init__(self, description: str, filename: str, as_image: bool, ts: float) -> None:
        super().__init__(ts)
        #: Attachment description.
        self.description: str = description
        #: Attachment filename, this is a path relative to the report directory.
        self.filename: str = filename
        #: Whether or not the attachment should be interpreted as an image (boolean).
        self.as_image: bool = as_image


class Url(StepLog):
    """
    The log resulting of log_url function.
    Inherits :py:class:`StepLog <lemoncheesecake.reporting.StepLog>`.
    """
    def __init__(self, description: str, url: str, ts: float) -> None:
        super().__init__(ts)
        #: Optional description.
        self.description: str = description
        #: Actual url.
        self.url: str = url


class Step:
    """
    This class holds logs occurring within a step.
    """
    def __init__(self, description: str) -> None:
        #: Step description.
        self.description: str = description
        #: Parent result.
        self.parent_result: Optional[Result] = None
        self._logs: List[StepLog] = []
        #: Step start time.
        self.start_time: Optional[float] = None
        #: Step end time.
        self.end_time: Optional[float] = None

    def add_log(self, log: StepLog) -> None:
        """
        Add a log to the step.
        """
        log.parent_step = self
        self._logs.append(log)

    def get_logs(self) -> List[StepLog]:
        """
        Get step logs.
        """
        return self._logs

    @staticmethod
    def _is_log_successful(log):
        if isinstance(log, Check):
            return log.is_successful
        elif isinstance(log, Log):
            return log.level != Log.LEVEL_ERROR
        return True

    def is_successful(self) -> bool:
        """
        Return whether or not (as a boolean) the step is successful.
        """
        return all(map(Step._is_log_successful, self._logs))

    @property
    def duration(self) -> Optional[float]:
        """
        Return the duration as a float or None if the step is not yet complete.
        """
        return _get_duration(self.start_time, self.end_time)


class Result:
    """
    Holds the result of setup/teardown phase and it's also the base class of
    :py:class:`TestResult <lemoncheesecake.reporting.TestResult>`.
    """

    #: possible status values for a Result instance
    STATUSES = "passed", "failed", "skipped", "disabled"

    def __init__(self) -> None:
        # please note that this attribute is also defined in TestResult through BaseTest,
        # one will override the other
        #: Parent suite.
        self.parent_suite: Optional[SuiteResult] = None
        #: Result type (it is one of the following: "test_session_setup", "test_session_teardown",
        #: "suite_setup", "suite_teardown", "test").
        self.type: Optional[str] = None
        self._steps: List[Step] = []
        #: Result start time.
        self.start_time: Optional[float] = None
        #: Result end time.
        self.end_time: Optional[float] = None
        #: Result status (one of Result.STATUSES or None if the result is not yet complete).
        self.status: Optional[str] = None
        #: Result status details, if any.
        self.status_details: Optional[str] = None

    def add_step(self, step: Step) -> None:
        """
        Add step to the result.
        """
        step.parent_result = self
        self._steps.append(step)

    def get_steps(self) -> List[Step]:
        """
        Get steps.
        """
        return self._steps

    def is_successful(self) -> bool:
        """
        Return whether or not the result is successful (even if the result is not yet complete).
        """
        if self.status:  # test is finished
            return self.status in ("passed", "disabled")
        else:  # check if the test is successful "so far"
            return all(step.is_successful() for step in self._steps)

    @property
    def duration(self) -> Optional[float]:
        """
        Return the duration as a float or None if the result is not yet complete.
        """
        return _get_duration(self.start_time, self.end_time)


class TestResult(BaseTest, Result):
    """
    Holds the result of a test. Inherits :py:class:`Result <lemoncheesecake.reporting.Result>`.

    :var str ~.name: test name
    :var str ~.description: test description
    :var list ~.tags: test tags
    :var dict ~.properties: test properties, as a dict
    :var dict ~.links: test links, as a list of tuples (`(url, description)` where description can be `None`)
    """
    def __init__(self, name: str, description: str) -> None:
        BaseTest.__init__(self, name, description)
        Result.__init__(self)
        self.type = "test"
        # non-serialized attributes (only set in-memory during test execution):
        self.rank = 0


class SuiteResult(BaseSuite):
    """
    Contains the results of tests and sub-suites within the suite.

    :var str ~.name: test name
    :var str ~.description: test description
    :var list ~.tags: test tags
    :var dict ~.properties: test properties, as a dict
    :var dict ~.links: test links, as a list of tuples (`(url, description)` where description can be `None`)
    """
    def __init__(self, name: str, description: str) -> None:
        BaseSuite.__init__(self, name, description)
        #: The suite start time.
        self.start_time: Optional[float] = None
        #: The suite end time.
        self.end_time: Optional[float] = None
        self._suite_setup: Optional[Result] = None
        self._suite_teardown: Optional[Result] = None
        # non-serialized attributes (only set in-memory during test execution)
        self.rank = 0

    @property
    def suite_setup(self) -> Result:
        """
        The suite setup result (if any).
        """
        return self._suite_setup
    
    @suite_setup.setter
    def suite_setup(self, setup: Result) -> None:
        if setup:
            setup.parent_suite = self
            setup.type = "suite_setup"
        self._suite_setup = setup

    @property
    def suite_teardown(self) -> Result:
        """
        The suite teardown result (if any).
        """
        return self._suite_teardown

    @suite_teardown.setter
    def suite_teardown(self, teardown: Result) -> None:
        if teardown:
            teardown.parent_suite = self
            teardown.type = "suite_teardown"
        self._suite_teardown = teardown

    @property
    def duration(self) -> Optional[float]:
        """
        The suite duration, which is the addition of all results contained in the suite and its sub-suite (recursively).
        """
        return reduce(
            lambda x, y: x + y,
            # result.duration is None if the corresponding result is in progress
            [result.duration or 0 for result in flatten_results([self])],
            0
        )

    def get_tests(self) -> List[TestResult]:
        """
        Return the tests contained in the suite.
        """
        tests = super().get_tests()
        return sorted(tests, key=lambda t: t.rank)

    def get_suites(self) -> List[SuiteResult]:
        """
        Return the sub-suites contained in the suite.
        """
        suites = super().get_suites()
        return sorted(suites, key=lambda s: s.rank)

    def pull_node(self):
        node = BaseSuite.pull_node(self)
        node._suite_setup = None
        node._suite_teardown = None
        return node

    def is_empty(self) -> bool:
        return BaseSuite.is_empty(self) and self._suite_setup is None and self._suite_teardown is None

    def filter(self, result_filter: Callable[[Result], bool]) -> SuiteResult:
        suite = BaseSuite.filter(self, result_filter)
        if self._suite_setup and result_filter(self._suite_setup):
            suite._suite_setup = self._suite_setup
        if self._suite_teardown and result_filter(self._suite_teardown):
            suite._suite_teardown = self._suite_teardown
        return suite


class ReportLocation:
    _TEST_SESSION_SETUP = 0
    _TEST_SESSION_TEARDOWN = 1
    _SUITE_SETUP = 2
    _SUITE_TEARDOWN = 3
    _TEST = 4

    def __init__(self, node_type, node_hierarchy=None):
        self.node_type = node_type
        self.node_hierarchy = node_hierarchy

    @classmethod
    def in_test_session_setup(cls) -> ReportLocation:
        return cls(cls._TEST_SESSION_SETUP)

    @classmethod
    def in_test_session_teardown(cls) -> ReportLocation:
        return cls(cls._TEST_SESSION_TEARDOWN)

    @classmethod
    def in_suite_setup(cls, suite: TreeNodeHierarchy) -> ReportLocation:
        return cls(cls._SUITE_SETUP, normalize_node_hierarchy(suite))

    @classmethod
    def in_suite_teardown(cls, suite: TreeNodeHierarchy) -> ReportLocation:
        return cls(cls._SUITE_TEARDOWN, normalize_node_hierarchy(suite))

    @classmethod
    def in_test(cls, test: TreeNodeHierarchy) -> ReportLocation:
        return cls(cls._TEST, normalize_node_hierarchy(test))

    def get(self, report: Report) -> Union[Result, SuiteResult, TestResult, None]:
        if self.node_type == self._TEST_SESSION_SETUP:
            return report.test_session_setup
        elif self.node_type == self._TEST_SESSION_TEARDOWN:
            return report.test_session_teardown
        elif self.node_type == self._SUITE_SETUP:
            suite = report.get_suite(self.node_hierarchy)
            return suite.suite_setup
        elif self.node_type == self._SUITE_TEARDOWN:
            suite = report.get_suite(self.node_hierarchy)
            return suite.suite_teardown
        elif self.node_type == self._TEST:
            return report.get_test(self.node_hierarchy)
        else:
            raise Exception("Unknown self type %s" % self.node_type)

    def __eq__(self, other):
        return all((
            isinstance(other, ReportLocation),
            self.node_type == other.node_type,
            self.node_hierarchy == other.node_hierarchy
        ))

    def __hash__(self):
        return hash((self.node_type, self.node_hierarchy))

    def __str__(self):
        ret = ""
        if self.node_hierarchy:
            ret += "'%s' " % ".".join(self.node_hierarchy)
        ret += ("session setup", "session teardown", "suite setup", "suite teardown", "test")[self.node_type]

        return "<ReportLocation %s>" % ret


def flatten_results(suites: Iterable[SuiteResult]) -> Iterator[Result]:
    for suite in flatten_suites(suites):
        if suite.suite_setup:
            yield suite.suite_setup
        yield from suite.get_tests()
        if suite.suite_teardown:
            yield suite.suite_teardown


class Report:
    DEFAULT_TITLE = "Test Report"

    def __init__(self) -> None:
        self.info = []
        self._test_session_setup: Optional[Result] = None
        self._test_session_teardown: Optional[Result] = None
        self._suites: List[SuiteResult] = []
        #: The test run start time.
        self.start_time: Optional[float] = None
        #: The test run end time.
        self.end_time: Optional[float] = None
        #: The report saving time.
        self.saving_time: Optional[float] = None
        #: The report title.
        self.title = Report.DEFAULT_TITLE
        #: The number of threads used for the test run.
        self.nb_threads = 1
        # both attributes enable the report to be saved back if Report.bind() as been called
        self.backend = None
        self.path = None

    @property
    def test_session_setup(self) -> Optional[Result]:
        """
        The session setup result if any.
        """
        return self._test_session_setup

    @test_session_setup.setter
    def test_session_setup(self, setup: Result) -> None:
        if setup:
            setup.type = "test_session_setup"
        self._test_session_setup = setup

    @property
    def test_session_teardown(self) -> Optional[Result]:
        """
        The session teardown result if any.
        """
        return self._test_session_teardown

    @test_session_teardown.setter
    def test_session_teardown(self, teardown: Result) -> None:
        if teardown:
            teardown.type = "test_session_teardown"
        self._test_session_teardown = teardown

    @property
    def duration(self) -> Optional[float]:
        """
        The test run duration.
        """
        return _get_duration(self.start_time, self.end_time)

    @property
    def nb_tests(self) -> int:
        """
        The number of tests in the report.
        """
        return len(list(self.all_tests()))

    @property
    def parallelized(self) -> bool:
        """
        Whether or not the tests were parallelized.
        """
        return self.nb_threads > 1 and self.nb_tests > 1

    def add_info(self, name: str, value: str) -> None:
        """
        Add extra information (name, value) in the report.
        """
        self.info.append([name, value])
    
    def add_suite(self, suite: SuiteResult) -> None:
        """
        Add suite result to the report.
        """
        self._suites.append(suite)
    
    def get_suites(self) -> List[SuiteResult]:
        """
        Get suite results.
        """
        return sorted(self._suites, key=lambda s: s.rank)

    def get_suite(self, hierarchy: List[str]) -> SuiteResult:
        return find_suite(self._suites, hierarchy)

    def get_test(self, hierarchy: List[str]) -> TestResult:
        return find_test(self._suites, hierarchy)

    def get(self, location: ReportLocation) -> Union[Result, SuiteResult, TestResult, None]:
        return location.get(self)

    def is_successful(self) -> bool:
        """
        Return whether or not the test run is considered as successful.

        Please note that every result is taken into account, including tests but also setups and teardowns.
        """
        return all(result.status in ("passed", "disabled") for result in self.all_results())

    def all_suites(self) -> Iterator[SuiteResult]:
        """
        An iterator over all suite results contained in the report (recursive).
        """
        return flatten_suites(self._suites)

    def all_tests(self) -> Iterator[TestResult]:
        """
        An iterator over all test results contained in the report.
        """
        return flatten_tests(self._suites)

    def all_results(self) -> Iterator[Result]:
        """
        An iterator over all results (tests, setups, teardowns) contained in the report.
        """
        if self.test_session_setup:
            yield self.test_session_setup
        yield from flatten_results(self.get_suites())
        if self.test_session_teardown:
            yield self.test_session_teardown

    def all_steps(self) -> Iterator[Step]:
        """
        An iterator over all steps contained in the report.
        """
        for result in self.all_results():
            yield from result.get_steps()

    def build_message(self, template: str) -> str:
        """
        Build a message from a template that contains variable placeholders.

        Example: with template "Test results: {passed}/{enabled} passed ({passed_pct})" the method will return
        for instance "Test results: 1/2 passed (50%)".

        The following variables are available:

        - ``start_time``: the test run start time

        - ``end_time``: the test run end time

        - ``duration``: the test run duration

        - ``total``: the total number of tests (including disabled tests)

        - ``enabled``: the total number of tests (excluding disabled tests)

        - ``passed``: the number of passed tests

        - ``passed_pct``: the number of passed tests in percentage of enabled tests

        - ``failed``: the number of failed tests

        - ``failed_pct``: the number of failed tests in percentage of enabled tests

        - ``skipped``: the number of skipped tests

        - ``skipped_pct``: the number of skipped tests in percentage of enabled tests

        - ``disabled``: the number of disabled tests

        - ``disabled_pct``: the number of disabled tests in percentage of all tests
        """
        stats = ReportStats.from_report(self)
        variables = {
            name: func(self, stats) for name, func in _report_message_variables.items()
        }
        return template.format(**variables)

    def bind(self, backend: ReportSerializerMixin, path: str) -> None:
        self.backend = backend
        self.path = path

    def save(self):
        """
        Save the report.
        """
        assert self.backend and self.path, "Cannot save unbound report"
        self.backend.save_report(self.path, self)


class ReportStats:
    def __init__(self):
        self.tests_nb = 0
        self.tests_nb_by_status = {s: 0 for s in Result.STATUSES}
        self.duration = None
        self.duration_cumulative = 0

    @property
    def tests_enabled_nb(self):
        return sum((self.tests_nb_by_status["passed"], self.tests_nb_by_status["failed"], self.tests_nb_by_status["skipped"]))

    @property
    def successful_tests_percentage(self):
        return (float(self.tests_nb_by_status["passed"]) / self.tests_enabled_nb * 100) if self.tests_enabled_nb else 0

    @property
    def duration_cumulative_description(self):
        description = humanize_duration(self.duration_cumulative)
        if self.duration:
            description += " (parallelization speedup factor is %.1f)" % (float(self.duration_cumulative) / self.duration)
        return description

    @classmethod
    def from_results(cls, results: List[Result], duration: Optional[int]) -> ReportStats:
        stats = cls()

        stats.duration = duration

        stats.duration_cumulative = sum(result.duration or 0 for result in results)

        tests = list(filter(lambda r: isinstance(r, TestResult), results))

        stats.tests_nb = len(tests)

        for test in tests:
            if test.status:
                stats.tests_nb_by_status[test.status] += 1

        return stats

    @classmethod
    def from_report(cls, report: Report) -> ReportStats:
        return cls.from_results(list(report.all_results()), report.duration)

    @classmethod
    def from_suites(cls, suites: List[SuiteResult], parallelized: bool) -> ReportStats:
        results = list(flatten_results(suites))

        return cls.from_results(
            results,
            results[-1].end_time - results[0].start_time if not parallelized else None
        )


def _percent(val, of):
    return "%d%%" % ((float(val) / of * 100) if of else 0)


_report_message_variables = {
    "start_time": lambda report, stats: time.asctime(time.localtime(report.start_time)),
    "end_time": lambda report, stats: time.asctime(time.localtime(report.end_time)),
    "duration": lambda report, stats: humanize_duration(report.end_time - report.start_time),

    "total": lambda report, stats: stats.tests_nb,
    "enabled": lambda report, stats: stats.tests_enabled_nb,

    "passed": lambda report, stats: stats.tests_nb_by_status["passed"],
    "passed_pct": lambda report, stats: _percent(stats.tests_nb_by_status["passed"], of=stats.tests_enabled_nb),

    "failed": lambda report, stats: stats.tests_nb_by_status["failed"],
    "failed_pct": lambda report, stats: _percent(stats.tests_nb_by_status["failed"], of=stats.tests_enabled_nb),

    "skipped": lambda report, stats: stats.tests_nb_by_status["skipped"],
    "skipped_pct": lambda report, stats: _percent(stats.tests_nb_by_status["skipped"], of=stats.tests_enabled_nb),

    "disabled": lambda report, stats: stats.tests_nb_by_status["disabled"],
    "disabled_pct": lambda report, stats: _percent(stats.tests_nb_by_status["disabled"], of=stats.tests_nb)
}


def check_report_message_template(template):
    try:
        template.format(**{name: "" for name in _report_message_variables})
        return template
    except KeyError as excp:
        raise ValueError("invalid report message template, unknown variable: %s" % excp)



'''
Created on Mar 28, 2017

@author: nicolas
'''

from typing import List, Any

from lemoncheesecake.helpers.orderedset import OrderedSet
from lemoncheesecake.helpers.text import jsonify
from lemoncheesecake.matching.matcher import Matcher, MatchResult, MatcherDescriptionTransformer


def _make_item(content, prefix="- "):
    lines = content.split("\n")
    return "\n".join(
        ["    " + prefix + lines[0]] +
        ["      " + line for line in lines[1:]]
    )


def _make_items(items, prefix="- ", relationship="and"):
    return [
        _make_item(item, prefix=prefix + relationship + " " if i > 0 else prefix)
            for i, item in enumerate(items)
    ]


def _build_multi_line_description(matchers, transformation, relationship):
    return "\n".join(
        [":"] +
        [
            _make_item(matcher.build_description(transformation), prefix="- %s " % relationship if i > 0 else "- ")
                for i, matcher in enumerate(matchers)
        ]
    )


def _build_single_line_description_if_suitable(matchers, transformation, relationship):
    if any(isinstance(matcher, (AllOf, AnyOf)) for matcher in matchers):
        # in case of "composite of composite", the multi-line rendering must be used in order to keep the logic
        return None

    descriptions = [matcher.build_description(transformation) for matcher in matchers]
    if any("\n" in desc for desc in descriptions):
        # in case of a '\n' in a matcher description, the resulting description will not be readable
        return None

    description = " {} ".format(relationship).join(descriptions)
    if len(description) > 100:
        # if the resulting description is more than 100 characters, we probably better keep multi-line rendering
        return None

    return description


def _build_composite_description(matchers, transformation, relationship):
    description = _build_single_line_description_if_suitable(matchers, transformation, relationship)
    if description:
        return description
    else:
        return _build_multi_line_description(matchers, transformation, relationship)


def _serialize_sub_matcher_result(matcher, result):
    content = "%s => %s" % (matcher.build_short_description(MatcherDescriptionTransformer()), "OK" if result else "KO")
    if result.description is not None:
        content += ", %s" % result.description
    return content


class AllOf(Matcher):
    def __init__(self, matchers: List[Matcher]) -> None:
        self.matchers = matchers

    def build_short_description(self, transformation):
        return ":"

    def build_description(self, transformation):
        return _build_composite_description(self.matchers, transformation, "and")

    def matches(self, actual):
        results = []
        is_success = True
        for matcher in self.matchers:
            result = matcher.matches(actual)
            results.append((matcher, result))
            if not result:
                is_success = False
                break

        match_details = "\n".join(
            ["got:"] +
            _make_items([
                _serialize_sub_matcher_result(matcher, result) for matcher, result in results
            ], relationship="and")
        )

        return MatchResult(is_success, match_details)


def all_of(*matchers: Any) -> Matcher:
    """Test if all matchers match (logical AND between matchers)."""
    return AllOf(list(map(is_, matchers)))


class AnyOf(Matcher):
    def __init__(self, matchers: List[Matcher]) -> None:
        self.matchers = matchers

    def build_short_description(self, transformation):
        return ":"

    def build_description(self, transformation):
        return _build_composite_description(self.matchers, transformation, "or")

    def matches(self, actual):
        results = []
        for matcher in self.matchers:
            match = matcher.matches(actual)
            if match:
                return match
            results.append(match)

        return MatchResult.failure(
            ", ".join(
                OrderedSet(result.description for result in results if result.description)
            )
        )


def any_of(*matchers: Any) -> Matcher:
    """Test if at least one of the matcher match (logical OR between matchers)"""
    return AnyOf(list(map(is_, matchers)))


class Anything(Matcher):
    def __init__(self, wording="to be anything"):
        self.wording = wording

    def build_description(self, transformation):
        return transformation(self.wording)

    def matches(self, actual):
        return MatchResult.success("got %s" % jsonify(actual))


def anything() -> Matcher:
    """Matches anything (always succeed, whatever the actual value)"""
    return Anything()


def something() -> Matcher:
    """Same thing as the 'anything' matcher but use 'to be something' in the matcher description"""
    return Anything(wording="to be something")


def existing() -> Matcher:
    """Same thing as the 'anything' matcher but use 'to exist' in the matcher description"""
    return Anything(wording="to exist")


def present() -> Matcher:
    """Same thing as the 'anything' matcher but use 'to be present' in the matcher description"""
    return Anything(wording="to be present")


def is_(matcher: Any) -> Matcher:
    """If the function argument is not an instance of Matcher, wrap it into
    a matcher using equal_to, otherwise return the matcher argument as-is"""
    from lemoncheesecake.matching.matchers.value import equal_to
    return matcher if isinstance(matcher, Matcher) else equal_to(matcher)


class Not(Matcher):
    def __init__(self, matcher: Matcher) -> None:
        self.matcher = matcher

    def build_description(self, transformation):
        transformation.negative = True
        return self.matcher.build_description(transformation)

    def matches(self, actual):
        result = self.matcher.matches(actual)
        return MatchResult(not result, result.description)


def not_(matcher: Any) -> Matcher:
    """Negates the matcher in argument"""
    return Not(is_(matcher))


is_not = not_



'''
Created on Mar 27, 2017

@author: nicolas
'''

from typing import Union, Any

from lemoncheesecake.helpers.text import jsonify
from lemoncheesecake.matching.matcher import Matcher, MatchResult, MatcherDescriptionTransformer
from lemoncheesecake.matching.matchers.composites import is_, not_
from lemoncheesecake.matching.matchers.types_ import is_bool


class EqualTo(Matcher):
    def __init__(self, expected):
        self.expected = expected

    def build_description(self, transformation):
        return transformation("to be equal to %s" % jsonify(self.expected))

    def matches(self, actual):
        from lemoncheesecake.matching import DISPLAY_DETAILS_WHEN_EQUAL

        if actual == self.expected:
            if DISPLAY_DETAILS_WHEN_EQUAL:
                return MatchResult.success("got %s" % jsonify(actual))
            else:
                return MatchResult.success()
        else:
            return MatchResult.failure("got %s" % jsonify(actual))


def equal_to(expected: Any) -> Matcher:
    """Test if value is equal to expected"""
    return EqualTo(expected)


class _Comparator(Matcher):
    def __init__(self, expected, comparison, comparison_description):
        self.expected = expected
        self.comparison = comparison
        self.comparison_description = comparison_description

    def build_description(self, transformation):
        return transformation("to be %s %s" % (self.comparison_description, jsonify(self.expected)))

    def matches(self, actual):
        return MatchResult(self.comparison(actual, self.expected), f"got {jsonify(actual)}")


def not_equal_to(expected: Any) -> Matcher:
    """
    Test if value is not equal to expected.
    """
    return _Comparator(expected, lambda a, e: a != e, "not equal to")


def greater_than(expected: Any) -> Matcher:
    """
    Test if value is greater than expected.
    """
    return _Comparator(expected, lambda a, e: a > e, "greater than")


def greater_than_or_equal_to(expected: Any) -> Matcher:
    """
    Test if value is greater than or equal to expected.
    """
    return _Comparator(expected, lambda a, e: a >= e, "greater than or equal to")


def less_than(expected: Any) -> Matcher:
    """
    Test if value is less than expected.
    """
    return _Comparator(expected, lambda a, e: a < e, "less than")


def less_than_or_equal_to(expected: Any) -> Matcher:
    """
    Test if value is less than or equal to expected.
    """
    return _Comparator(expected, lambda a, e: a <= e, "less than or equal to")


class IsBetween(Matcher):
    def __init__(self, min, max):
        self.min = min
        self.max = max

    def build_description(self, transformation):
        return transformation("to be between %s and %s" % (self.min, self.max))

    def matches(self, actual):
        return MatchResult(self.min <= actual <= self.max, "got %s" % jsonify(actual))


def is_between(min: Union[int, float], max: Union[int, float]) -> Matcher:
    """Test if value is between min and max"""
    return IsBetween(min, max)


class IsNone(Matcher):
    def build_description(self, transformation):
        return transformation("to be null")

    def matches(self, actual):
        if actual is None:
            return MatchResult.success("got null")
        else:
            return MatchResult.failure("got %s" % jsonify(actual))


def is_none() -> Matcher:
    """Test if value is None"""
    return IsNone()


def is_not_none() -> Matcher:
    """Test if value is not None"""
    return not_(is_none())


class HasLength(Matcher):
    def __init__(self, matcher: Matcher) -> None:
        self.matcher = matcher

    def build_description(self, transformation):
        return transformation(
            "to have a length that %s" % (self.matcher.build_description(MatcherDescriptionTransformer(conjugate=True)))
        )

    def matches(self, actual):
        return self.matcher.matches(len(actual))


def has_length(length: Union[int, Matcher]) -> Matcher:
    """Test if value has a length of"""
    return HasLength(is_(length))


def is_true() -> Matcher:
    """Test if value is true (boolean type)"""
    return is_bool(True)


def is_false() -> Matcher:
    """Test if value is false (boolean type)"""
    return is_bool(False)



'''
Created on Jun 14, 2017

@author: nicolas
'''

from decimal import Decimal
from functools import reduce

# This junit backend implementation is based on the documentation
# http://llg.cubic.org/docs/junit/
# also see:
# https://confluence.atlassian.com/display/BAMBOO/JUnit+parsing+in+Bamboo

try:
    from lxml import etree as ET
    from lxml.builder import E
    LXML_IS_AVAILABLE = True
except ImportError:
    LXML_IS_AVAILABLE = False

from lemoncheesecake.reporting.backend import FileReportBackend
from lemoncheesecake.reporting import ReportStats, Log, Check, format_time_as_iso8601
from lemoncheesecake.reporting.backends.xml import make_xml_child, make_xml_node, indent_xml, DEFAULT_INDENT_LEVEL


def _serialization_duration(duration):
    return "%d.%03d" % (int(duration), Decimal(round(duration, 3)) % 1 * 1000)


def _serialize_test_result(test):
    xml_test = make_xml_node(
        "testcase",
        "name", test.name,
        "time", _serialization_duration(test.duration or 0),
    )
    
    if test.status == "skipped":
        make_xml_child(xml_test, "skipped")
    else:
        for step in test.get_steps():
            for log in step.get_logs():
                if isinstance(log, Check) and log.is_successful is False:
                    make_xml_child(
                        xml_test, "failure",
                        "message", "failed check in step '%s', %s%s" % (
                            step.description, log.description,
                            (": " + log.details) if log.details else ""
                        )
                    )
                elif isinstance(log, Log) and log.level == Log.LEVEL_ERROR:
                    make_xml_child(
                        xml_test, "error",
                        "message", "error log in step '%s': %s" % (
                            step.description, log.message
                        )
                    )

    return xml_test


def _serialize_suite_result(suite):
    tests = suite.get_tests()
    xml_suite = make_xml_node(
        "testsuite",
        "name", suite.path,
        "tests", str(len(tests)),
        "failures", str(len(list(filter(lambda t: t.status == "failed", tests)))),
        "skipped", str(len(list(filter(lambda t: t.status == "skipped", tests)))),
        "time", _serialization_duration(reduce(lambda x, y: x + y, (t.duration or 0 for t in tests), 0)),
        "timestamp", format_time_as_iso8601(min(t.start_time for t in tests))
    )
    xml_suite.extend(map(_serialize_test_result, tests))
    return xml_suite


def serialize_report_as_xml_tree(report):
    xml_report = E("testsuites")

    stats = ReportStats.from_report(report)

    xml_report.attrib["tests"] = str(stats.tests_nb_by_status["passed"])
    xml_report.attrib["failures"] = str(stats.tests_nb_by_status["failed"])

    if report.end_time is not None:
        xml_report.attrib["time"] = _serialization_duration(report.end_time - report.start_time)

    for suite in report.all_suites():
        if suite.get_tests():
            xml_report.append(_serialize_suite_result(suite))

    return xml_report


def serialize_report_as_string(report, indent_level=DEFAULT_INDENT_LEVEL):
    xml_report = serialize_report_as_xml_tree(report)
    indent_xml(xml_report, indent_level=indent_level)

    return ET.tostring(xml_report, pretty_print=True, encoding="unicode")


def save_report_into_file(report, filename, indent_level=DEFAULT_INDENT_LEVEL):
    content = serialize_report_as_string(report, indent_level)
    with open(filename, "w") as fh:
        fh.write(content)


class JunitBackend(FileReportBackend):
    def __init__(self):
        self.indent_level = DEFAULT_INDENT_LEVEL

    def get_name(self):
        return "junit"

    def is_available(self):
        return LXML_IS_AVAILABLE

    def get_report_filename(self):
        return "report-junit.xml"

    def save_report(self, filename, report):
        save_report_into_file(report, filename, self.indent_level)



'''
Created on Mar 19, 2016

@author: nicolas
'''

import sys

from termcolor import colored

from lemoncheesecake.testtree import filter_suites, flatten_suites
from lemoncheesecake.reporting.backend import ReportingBackend, ReportingSession, ReportingSessionBuilderMixin
from lemoncheesecake.reporting.report import ReportStats
from lemoncheesecake.helpers.time import humanize_duration
from lemoncheesecake.helpers.text import ensure_single_line_text
from lemoncheesecake.helpers import terminalsize
from lemoncheesecake.reporting.console import test_status_to_color


class LinePrinter:
    def __init__(self, terminal_width):
        self.terminal_width = terminal_width
        self.prev_len = 0

    def print_line(self, line, force_len=None):
        value_len = force_len if force_len else len(line)
        if value_len >= self.terminal_width - 1:
            line = line[:self.terminal_width - 5] + "..."
            value_len = len(line)

        sys.stdout.write("\r")
        sys.stdout.write(line)
        if self.prev_len > value_len:
            sys.stdout.write(" " * (self.prev_len - value_len))
        sys.stdout.flush()

        self.prev_len = value_len

    def new_line(self):
        self.prev_len = 0
        sys.stdout.write("\n")
        sys.stdout.flush()

    def erase_line(self):
        sys.stdout.write("\r")
        sys.stdout.write(" " * self.prev_len)
        sys.stdout.write("\r")
        self.prev_len = 0


def _make_suite_header_line(suite, terminal_width):
    suite_name = suite.path
    max_width = min((terminal_width, 80))
    # -2 corresponds to the two space characters at the left and right of suite path + another character to avoid
    # an extra line after the suite line on Windows terminal having width <= 80
    padding_total = max_width - 3 - len(suite_name) if len(suite_name) <= (max_width - 3) else 0
    padding_left = padding_total // 2
    padding_right = padding_total // 2 + padding_total % 2

    return "=" * padding_left + " " + colored(suite_name, attrs=["bold"]) + " " + "=" * padding_right


def _make_test_status_label(status):
    if status == "passed":
        label = "OK"
    elif status in ("skipped", "disabled", None):
        label = "--"
    else:
        label = "KO"

    return colored(label, test_status_to_color(status), attrs=["bold"])


def _make_test_result_line(name, num, status):
    line = " %s %2s # %s" % (_make_test_status_label(status), num, name)
    raw_line = "%s %2s # %s" % ("OK" if status == "passed" else "KO", num, name)

    return line, len(raw_line)


def _print_summary(stats, parallel=False):
    print()
    print(colored("Statistics", attrs=["bold"]), ":")
    print(" * Duration: %s" % (humanize_duration(stats.duration) if stats.duration is not None else "n/a"))
    if parallel:
        print(" * Cumulative duration: %s" % stats.duration_cumulative_description)
    print(" * Tests: %d" % stats.tests_nb)
    print(" * Successes: %d (%d%%)" % (stats.tests_nb_by_status["passed"], stats.successful_tests_percentage))
    print(" * Failures: %d" % (stats.tests_nb_by_status["failed"]))
    if stats.tests_nb_by_status["skipped"]:
        print(" * Skipped: %d" % (stats.tests_nb_by_status["skipped"]))
    if stats.tests_nb_by_status["disabled"]:
        print(" * Disabled: %d" % (stats.tests_nb_by_status["disabled"]))
    print()


class SequentialConsoleReportingSession(ReportingSession):
    def __init__(self, terminal_width, show_test_full_path, report):
        self.terminal_width = terminal_width
        self.show_test_full_path = show_test_full_path
        self.report = report
        self.lp = LinePrinter(self.terminal_width)
        self.context = None
        self.custom_step_prefix = None
        self.current_suite = None

    def get_test_label(self, test):
        if self.show_test_full_path:
            return test.path
        return test.name

    def ensure_suite_header_is_displayed(self, suite):
        if suite == self.current_suite:
            return

        self.current_suite = suite
        self.current_test_idx = 1

        if self.previous_obj:
            sys.stdout.write("\n")

        sys.stdout.write(_make_suite_header_line(suite, self.terminal_width) + "\n")

        self.previous_obj = suite

    def on_test_session_start(self, event):
        self.previous_obj = None

    def on_suite_setup_start(self, event):
        self.ensure_suite_header_is_displayed(event.suite)

        self.step_prefix = " => setup suite: "
        self.lp.print_line(self.step_prefix + "...")

    def on_suite_teardown_start(self, event):
        self.step_prefix = " => teardown suite: "
        self.lp.print_line(self.step_prefix + "...")

    def on_test_session_setup_start(self, event):
        self.step_prefix = " => setup test session: "
        self.lp.print_line(self.step_prefix + "...")

    def on_test_session_teardown_start(self, event):
        self.step_prefix = " => teardown test session: "
        self.lp.print_line(self.step_prefix + "...")

    def on_suite_setup_end(self, event):
        self.lp.erase_line()
        self.custom_step_prefix = None

    on_suite_teardown_end = on_suite_setup_end

    def on_test_session_setup_end(self, event):
        self.lp.erase_line()
        self.custom_step_prefix = None

    on_test_session_teardown_end = on_test_session_setup_end

    def on_test_start(self, event):
        self.ensure_suite_header_is_displayed(event.test.parent_suite)

        self.step_prefix = " -- %2s # %s" % (self.current_test_idx, self.get_test_label(event.test))
        self.lp.print_line(self.step_prefix + "...")
        self.previous_obj = event.test

    def on_test_end(self, event):
        test_data = self.report.get_test(event.test)

        line, raw_line_len = _make_test_result_line(
            self.get_test_label(event.test), self.current_test_idx, test_data.status
        )

        self.lp.print_line(line, force_len=raw_line_len)
        self.lp.new_line()

        self.current_test_idx += 1

    def _bypass_test(self, test, status):
        self.ensure_suite_header_is_displayed(test.parent_suite)

        line = " %s %2s # %s" % (_make_test_status_label(status), self.current_test_idx, self.get_test_label(test))
        raw_line = "%s %2s # %s" % ("KO", self.current_test_idx, self.get_test_label(test))
        self.lp.print_line(line, force_len=len(raw_line))
        self.lp.new_line()
        self.current_test_idx += 1

    def on_test_skipped(self, event):
        self._bypass_test(event.test, "skipped")

    def on_test_disabled(self, event):
        self._bypass_test(event.test, "disabled")

    def on_step_start(self, event):
        self.lp.print_line("%s (%s...)" % (self.step_prefix, ensure_single_line_text(event.step_description)))

    def on_test_session_end(self, event):
        _print_summary(ReportStats.from_report(self.report), self.report.parallelized)


class ParallelConsoleReportingSession(ReportingSession):
    def __init__(self, terminal_width, report):
        self.terminal_width = terminal_width
        self.report = report
        self.lp = LinePrinter(self.terminal_width)
        self.current_test_idx = 1

    def on_test_end(self, event):
        test_data = self.report.get_test(event.test)

        line, _ = _make_test_result_line(
            event.test.path, self.current_test_idx, test_data.status
        )

        print(line)

        self.current_test_idx += 1

    def _bypass_test(self, test, status):
        line = " %s %2s # %s" % (_make_test_status_label(status), self.current_test_idx, test.path)
        print(line)
        self.current_test_idx += 1

    def on_test_skipped(self, event):
        self._bypass_test(event.test, "skipped")

    def on_test_disabled(self, event):
        self._bypass_test(event.test, "disabled")

    def on_test_session_end(self, event):
        _print_summary(ReportStats.from_report(self.report), self.report.parallelized)


class ConsoleBackend(ReportingBackend, ReportingSessionBuilderMixin):
    def __init__(self):
        width, height = terminalsize.get_terminal_size()
        self.terminal_width = width
        self.show_test_full_path = True

    def get_name(self):
        return "console"

    def create_reporting_session(self, report_dir, report, parallel, saving_strategy):
        return \
            ParallelConsoleReportingSession(self.terminal_width, report) if parallel else \
            SequentialConsoleReportingSession(self.terminal_width, self.show_test_full_path, report)


def print_report_as_test_run(report, test_filter):
    suites = filter_suites(report.get_suites(), test_filter)

    ###
    # Setup terminal
    ###
    terminal_width, _ = terminalsize.get_terminal_size()

    ###
    # Display suite results
    ###
    suite_idx = 0
    for suite in flatten_suites(suites):
        if len(suite.get_tests()) == 0:
            continue
        if suite_idx > 0:
            print()
        header_line = _make_suite_header_line(suite, terminal_width)
        print(header_line)
        for test_idx, test in enumerate(suite.get_tests()):
            test_result_line, _ = _make_test_result_line(test.path, num=test_idx+1, status=test.status)
            print(test_result_line)
        suite_idx += 1

    ###
    # Display summary
    ###
    if suite_idx > 0:
        if test_filter:
            stats = ReportStats.from_suites(suites, report.parallelized)
        else:
            stats = ReportStats.from_report(report)
        _print_summary(stats, report.parallelized)
    else:
        print("No test found or no matching test in the report")



import sys
import time
import re
import inspect
import threading
from contextlib import contextmanager

from queue import Queue

from lemoncheesecake.helpers.text import camel_case_to_snake_case
from lemoncheesecake.exceptions import serialize_current_exception

DEBUG = False


class Event:
    def __init__(self, event_time=None):
        self.time = event_time or time.time()

    @classmethod
    def get_name(cls):
        return re.sub(r"_event$", "", camel_case_to_snake_case(cls.__name__))

    def __str__(self):
        return "<Event type='%s'>" % self.get_name()


class EventType:
    def __init__(self, event_class):
        self._event_class = event_class
        self._handlers = []

    def subscribe(self, handler):
        self._handlers.append(handler)

    def unsubscribe(self, handler):
        self._handlers.remove(handler)

    def reset(self):
        self._handlers = []

    def handle(self, event):
        for handler in self._handlers:
            handler(event)


class EventManager:
    def __init__(self):
        self._event_types = {}

    @staticmethod
    def _get_event_classes():
        """
        Get available event classes by introspecting this module to get available event classes.
        """
        current_module = sys.modules[__name__]
        for sym_name in dir(current_module):
            if sym_name.startswith("_"):
                continue
            sym = getattr(current_module, sym_name)
            if inspect.isclass(sym) and sym is not Event and issubclass(sym, Event):
                yield sym

    @classmethod
    def load(cls):
        eventmgr = cls()
        for event_class in eventmgr._get_event_classes():
            eventmgr.register_event(event_class)
        return eventmgr

    @staticmethod
    def _get_event_name(val):
        # val can be either an Event-based class or the name of the event itself (as an str)
        return val.get_name() if inspect.isclass(val) and issubclass(val, Event) else val

    def register_event(self, *event_classes):
        for event_class in event_classes:
            self._event_types[event_class.get_name()] = EventType(event_class)

    def subscribe_to_event(self, event, handler):
        self._event_types[self._get_event_name(event)].subscribe(handler)

    def subscribe_to_events(self, handlers):
        for event, handler in handlers.items():
            self.subscribe_to_event(event, handler)

    def add_listener(self, listener):
        for event_name in self._event_types:
            handler_name = "on_%s" % event_name
            handler = getattr(listener, handler_name, None)
            if handler and callable(handler):
                self.subscribe_to_event(event_name, handler)

    def unsubscribe_from_event(self, event, handler):
        self._event_types[self._get_event_name(event)].unsubscribe(handler)

    def handle_event(self, event):
        self._event_types[event.__class__.get_name()].handle(event)

    def fire(self, event):
        raise NotImplementedError()


class AsyncEventManager(EventManager):
    def __init__(self):
        EventManager.__init__(self)
        self._queue = None
        self._pending_failure = None, None

    def fire(self, event):
        assert self._queue, "Events can't be fired outside the 'handle_events' context manager."
        if DEBUG:
            print("Fire event %s" % event)
        self._queue.put(event)

    def get_pending_failure(self):
        return self._pending_failure

    def _handler_loop(self):
        while True:
            event = self._queue.get()
            if event is None:
                break
            try:
                self.handle_event(event)
            except Exception as excp:
                self._pending_failure = excp, serialize_current_exception()
                break
            finally:
                self._queue.task_done()

    @contextmanager
    def handle_events(self):
        self._queue = Queue()

        thread = threading.Thread(target=self._handler_loop)
        thread.start()

        try:
            yield
        finally:
            self._queue.put(None)
            thread.join()
            self._queue = None


class SyncEventManager(EventManager):
    def fire(self, event):
        if DEBUG:
            print("Fire event %s" % event)
        return self.handle_event(event)


###
# Events related to the test session
###

class _ReportEvent(Event):
    def __init__(self, report, event_time=None):
        super().__init__(event_time)
        self.report = report


class TestSessionStartEvent(_ReportEvent):
    pass


class TestSessionEndEvent(_ReportEvent):
    pass


class TestSessionSetupStartEvent(Event):
    pass


class TestSessionSetupEndEvent(Event):
    pass


class TestSessionTeardownStartEvent(Event):
    pass


class TestSessionTeardownEndEvent(Event):
    pass


###
# Suite events
###

class _SuiteEvent(Event):
    def __init__(self, suite, event_time=None):
        super().__init__(event_time)
        self.suite = suite

    def __str__(self):
        return "<Event type='%s' suite='%s'>" % (self.get_name(), self.suite.path)


class SuiteStartEvent(_SuiteEvent):
    pass


class SuiteEndEvent(_SuiteEvent):
    pass


class SuiteSetupStartEvent(_SuiteEvent):
    pass


class SuiteSetupEndEvent(_SuiteEvent):
    pass


class SuiteTeardownStartEvent(_SuiteEvent):
    pass


class SuiteTeardownEndEvent(_SuiteEvent):
    pass


###
# Test events
###

class _TestEvent(Event):
    def __init__(self, test, event_time=None):
        super().__init__(event_time)
        self.test = test

    def __str__(self):
        return "<Event type='%s' test='%s'>" % (self.get_name(), self.test.path)


class TestStartEvent(_TestEvent):
    pass


class TestEndEvent(_TestEvent):
    pass


class TestSkippedEvent(_TestEvent):
    def __init__(self, test, reason, event_time=None):
        super().__init__(test, event_time)
        self.skipped_reason = reason


class TestDisabledEvent(_TestEvent):
    def __init__(self, test, reason, event_time=None):
        super().__init__(test, event_time)
        self.disabled_reason = reason


###
# Transverse test execution events
###

class RuntimeEvent(Event):
    def __init__(self, location, event_time=None):
        super().__init__(event_time)
        self.location = location


class StepStartEvent(RuntimeEvent):
    def __init__(self, location, description, thread_id, event_time=None):
        super().__init__(location, event_time)
        self.step_description = description
        self.thread_id = thread_id

    def __str__(self):
        return "<Event type='%s' description='%s'>" % (
            self.get_name(), self.step_description
        )


class StepEndEvent(RuntimeEvent):
    def __init__(self, location, step, thread_id, event_time=None):
        super().__init__(location, event_time)
        self.step = step
        self.thread_id = thread_id


class SteppedEvent(RuntimeEvent):
    """
    This event class cannot be instantiated directly and only serve has a base
    class for all events happening within a step.
    """
    def __init__(self, location, step, thread_id, event_time=None):
        super().__init__(location, event_time)
        self.step = step
        self.thread_id = thread_id


class LogEvent(SteppedEvent):
    def __init__(self, location, step, thread_id, level, message, event_time=None):
        super().__init__(location, step, thread_id, event_time)
        self.log_level = level
        self.log_message = message

    def __str__(self):
        return "<Event type='%s' level='%s' message='%s'>" % (
            self.get_name(), self.log_level, self.log_message
        )


class CheckEvent(SteppedEvent):
    def __init__(self, location, step, thread_id, description, is_successful, details=None, event_time=None):
        super().__init__(location, step, thread_id, event_time)
        self.check_description = description
        self.check_is_successful = is_successful
        self.check_details = details

    def __str__(self):
        return "<Event type='%s' description='%s' details='%s' is_successful='%s'>" % (
            self.get_name(), self.check_description, self.check_details,
            "success" if self.check_is_successful else "failure"
        )


class LogAttachmentEvent(SteppedEvent):
    def __init__(self, location, step, thread_id, path, description, as_image, event_time=None):
        super().__init__(location, step, thread_id, event_time)
        self.attachment_path = path
        self.attachment_description = description
        self.as_image = as_image

    def __str__(self):
        return "<Event type='%s' path='%s' description='%s'>" % (
            self.get_name(), self.attachment_path, self.attachment_description
        )


class LogUrlEvent(SteppedEvent):
    def __init__(self, location, step, thread_id, url, description, event_time=None):
        super().__init__(location, step, thread_id, event_time)
        self.url = url
        self.url_description = description

    def __str__(self):
        return "<Event type='%s' url='%s' description='%s'>" % (
            self.get_name(), self.url, self.url_description
        )



from terminaltables import AsciiTable
from termcolor import colored


def print_table(title, headers, lines):
    if lines:
        print("%s:" % title)
        print(AsciiTable([headers] + lines).table)
    else:
        print("%s: <none>" % title)
    print()


def bold(s):
    return colored(s, attrs=["bold"])



import os
import os.path as osp
from shutil import copy, copytree

from lemoncheesecake.helpers.resources import get_resource_path
from lemoncheesecake.reporting.backend import ReportingBackend, ReportingSessionBuilderMixin, ReportingSession
from lemoncheesecake.helpers.console import bold


class HtmlReportWriter(ReportingSession):
    def __init__(self, report_dir):
        self.report_dir = report_dir

    def on_test_session_start(self, _):
        src_dir = get_resource_path("html")
        resources_dir = osp.join(self.report_dir, ".html")

        os.mkdir(resources_dir)
        copy(osp.join(src_dir, "report.js"), resources_dir)
        copy(osp.join(src_dir, "report.css"), resources_dir)
        copy(osp.join(src_dir, "logo.png"), resources_dir)
        copy(osp.join(src_dir, "bootstrap-icons.css"), resources_dir)
        os.mkdir(osp.join(resources_dir, "fonts"))
        copy(osp.join(src_dir, "fonts", "bootstrap-icons.woff"), osp.join(resources_dir, "fonts"))
        copy(osp.join(src_dir, "fonts", "bootstrap-icons.woff2"), osp.join(resources_dir, "fonts"))
        copy(osp.join(src_dir, "report.html"), self.report_dir)

    def on_test_session_end(self, _):
        print("%s : file://%s/report.html" % (bold("HTML report"), self.report_dir))
        print()


class HtmlBackend(ReportingBackend, ReportingSessionBuilderMixin):
    def get_name(self):
        return "html"

    def create_reporting_session(self, report_dir, report, parallel, saving_strategy):
        return HtmlReportWriter(report_dir)



import inspect


def object_has_method(obj, method_name):
    try:
        return callable(getattr(obj, method_name))
    except AttributeError:
        return False


def get_callable_args(cb):
    if not inspect.isroutine(cb):
        try:
            cb = getattr(cb, "__call__")
        except AttributeError:
            raise ValueError("%s is not something that can be called" % cb)

    spec = inspect.getfullargspec(cb)
    args_start = 1 if inspect.ismethod(cb) else 0
    return spec.args[args_start:]


# adapted from https://stackoverflow.com/a/3681323
def _get_unbound_object_attr(obj, attr_name):
    for cls in inspect.getmro(obj.__class__):
        if attr_name in cls.__dict__:
            return cls.__dict__[attr_name]
    raise AttributeError("Cannot find attribute '%s' in %s" % (attr_name, obj))


def _is_property(obj, attr_name):
    try:
        attr = _get_unbound_object_attr(obj, attr_name)
    except AttributeError:
        # there is no such attribute defined in the class, meaning it must be an instance attribute
        # then not a property
        return False
    return isinstance(attr, property)


def _get_class_object_attributes(obj):
    for attr_name in dir(obj):
        if not any((attr_name.startswith("__"), _is_property(obj, attr_name))):
            yield attr_name, getattr(obj, attr_name)


def _get_module_object_attributes(mod):
    return ((attr_name, getattr(mod, attr_name)) for attr_name in dir(mod))


def get_object_attributes(obj):
    return _get_module_object_attributes(obj) if inspect.ismodule(obj) else _get_class_object_attributes(obj)



'''
Created on Sep 8, 2016

@author: nicolas
'''

from lemoncheesecake.exceptions import ValidationError
from lemoncheesecake.testtree import flatten_suites


class MetadataPolicy:
    def __init__(self):
        self._properties = {}
        self._tags = {}
        self._disallow_unknown_properties = False
        self._disallow_unknown_tags = False

    def has_constraints(self):
        return self._properties or self._tags or self._disallow_unknown_properties or self._disallow_unknown_tags

    @staticmethod
    def _get_rule_application(on_test, on_suite):
        if on_test is None and on_suite is None:
            return True, False
        if not on_test and not on_suite:
            raise AssertionError("either on_test or on_suite need to be True")
        return bool(on_test), bool(on_suite)

    def add_property_rule(self, prop_name, accepted_values=None, on_test=None, on_suite=None, required=False):
        """
        Declare a property rule.

        :param prop_name: the property name
        :param accepted_values: an optional list of accepted values
        :param on_test: whether or not the property can be used on a test
        :param on_suite: whether or not the property can be used on a suite
        :param required: whether or not the property is required

        If neither on_test or on_suite argument are set, then the property is only available for tests.
        """
        on_test, on_suite = self._get_rule_application(on_test, on_suite)
        self._properties[prop_name] = {
            "values": accepted_values,
            "on_test": on_test,
            "on_suite": on_suite,
            "required": required
        }

    def disallow_unknown_properties(self):
        """
        Disallow unknown properties for tests and suites.
        """
        self._disallow_unknown_properties = True

    def add_tag_rule(self, tag_name, on_test=None, on_suite=None):
        """
        Declare a tag rule.

        :param tag_name: the tag name
        :param on_test: whether or not the tag can be used on a test
        :param on_suite: whether or not the tag can be used on a suite

        If neither on_test or on_suite argument are set, then the tag is only available for tests.
        """
        tag_names = tag_name if type(tag_name) in (list, tuple) else [tag_name]
        for tag_name in tag_names:
            on_test, on_suite = self._get_rule_application(on_test, on_suite)
            self._tags[tag_name] = {
                "on_test": on_test,
                "on_suite": on_suite
            }

    def disallow_unknown_tags(self):
        """
        Disallow unknown tags for tests and suites.
        """
        self._disallow_unknown_tags = True

    def _check_compliance(self, obj, obj_type,
                          available_properties, forbidden_properties,
                          available_tags, forbidden_tags):
        # check unknown properties
        if self._disallow_unknown_properties:
            for property_name in obj.properties.keys():
                if property_name not in available_properties:
                    help_msg = "available are %s" % ", ".join(map(repr, available_properties.keys())) \
                        if available_properties else "no property is available"
                    raise ValidationError(
                        "In %s '%s', the property '%s' is not allowed (%s)" % (
                            obj_type, obj.path, property_name, help_msg
                        )
                    )

        # check forbidden properties
        for property_name in obj.properties.keys():
            if property_name in forbidden_properties:
                raise ValidationError(
                    "In %s '%s', the property '%s' is not allowed on a %s" % (
                        obj_type, obj.path, property_name, obj_type
                    )
                )

        # check required properties
        for required_property in filter(lambda p: available_properties[p]["required"], available_properties.keys()):
            if required_property not in obj.properties.keys():
                raise ValidationError(
                    "In %s '%s', the mandatory property '%s' is missing" % (
                        obj_type, obj.path, required_property
                    )
                )

        # check properties allowed values
        for name, value in obj.properties.items():
            if name not in available_properties:
                continue
            if available_properties[name]["values"] and value not in available_properties[name]["values"]:
                raise ValidationError(
                    "In %s '%s', value '%s' of property '%s' is not among accepted values: %s" % (
                        obj_type, obj.path, value, name, available_properties[name]["values"]
                    )
                )

        # check unknown tags
        if self._disallow_unknown_tags:
            for tag in obj.tags:
                if tag not in available_tags.keys():
                    help_msg = "available are %s" % ", ".join(map(repr, available_tags.keys())) \
                        if available_tags else "no tag is available"
                    raise ValidationError(
                        "In %s '%s', the tag '%s' is not allowed (%s)" % (
                            obj_type, obj.path, tag, help_msg
                        )
                    )

        # check forbidden tags
        for tag in obj.tags:
            if tag in forbidden_tags:
                raise ValidationError(
                    "In %s '%s', the tag '%s' is not allowed on a %s" % (
                        obj_type, obj.path, tag, obj_type
                    )
                )

    def check_test_compliance(self, test):
        """
        Check if the test complies to the metadata policy.
        Raise ValidationError if not compliant.
        """
        self._check_compliance(
            test, "test",
            {prop_name: p for prop_name, p in self._properties.items() if p["on_test"]},
            [prop_name for prop_name, p in self._properties.items() if not p["on_test"]],
            {tag_name: t for tag_name, t in self._tags.items() if t["on_test"]},
            [tag_name for tag_name, t in self._tags.items() if not t["on_test"]]
        )

    def check_suite_compliance(self, suite):
        """
        Check if the suite complies to the metadata policy.
        If recursive is set to True (which is the default), then suite tests and sub suites are also checked.
        Raise ValidationError if not compliant.
        """
        self._check_compliance(
            suite, "suite",
            {prop_name: p for prop_name, p in self._properties.items() if p["on_suite"]},
            [prop_name for prop_name, p in self._properties.items() if not p["on_suite"]],
            {tag_name: t for tag_name, t in self._tags.items() if t["on_suite"]},
            [tag_name for tag_name, t in self._tags.items() if not t["on_suite"]]
        )

        for test in suite.get_tests():
            self.check_test_compliance(test)

    def check_suites_compliance(self, suites):
        """
        Check if the suites comply to the metadata policy.
        Raise ValidationError if not compliant.
        """
        for suite in flatten_suites(suites):
            self.check_suite_compliance(suite)



import sys
import traceback


class LemoncheesecakeException(Exception):
    pass


class ProjectLoadingError(LemoncheesecakeException):
    pass


class ProjectNotFound(ProjectLoadingError):
    pass


class ModuleImportError(LemoncheesecakeException):
    pass


class ValidationError(LemoncheesecakeException):
    pass


class SuiteLoadingError(LemoncheesecakeException):
    pass


class FixtureLoadingError(LemoncheesecakeException):
    pass


class AbortTest(LemoncheesecakeException):
    """
    Raising this exception will stop the currently running test.
    """
    # NB: sphinx requires the constructor to be overridden, otherwise it raises an error
    def __init__(self, *args):
        LemoncheesecakeException.__init__(self, *args)


class AbortSuite(LemoncheesecakeException):
    """
    Raising this exception will stop the currently running suite.
    """
    # NB: sphinx requires the constructor to be overridden, otherwise it raises an error
    def __init__(self, *args):
        LemoncheesecakeException.__init__(self, *args)


class AbortAllTests(LemoncheesecakeException):
    """
    Raising this exception will stop the currently running test and all the tests waiting to be run.
    """
    # NB: sphinx requires the constructor to be overridden, otherwise it raises an error
    def __init__(self, *args):
        LemoncheesecakeException.__init__(self, *args)


class UserError(LemoncheesecakeException):
    """
    This exception is intended to be raised in pre-run and post-run phases of the project
    to indicate that a required state has not been fulfilled.
    """
    # NB: sphinx requires the constructor to be overridden, otherwise it raises an error
    def __init__(self, *args):
        LemoncheesecakeException.__init__(self, *args)


class ReportLoadingError(LemoncheesecakeException):
    pass


class TaskFailure(LemoncheesecakeException):
    pass


def serialize_current_exception(show_stacktrace=True):
    if show_stacktrace:
        return "\n" + "<" * 72 + "\n" + traceback.format_exc() + ">" * 72
    else:
        return " " + str(sys.exc_info()[1])



'''
Created on Dec 10, 2016

@author: nicolas
'''

import sys
import os
import os.path as osp
import shutil
import argparse
import inspect

from typing import Any, Dict, Optional, Sequence, Tuple, List

from lemoncheesecake.session import Session
from lemoncheesecake.events import AsyncEventManager
from lemoncheesecake.suite import load_suites_from_directory, Suite, resolve_tests_dependencies
from lemoncheesecake.runner import run_suites
from lemoncheesecake.fixture import load_fixtures_from_directory, Fixture, FixtureRegistry, BuiltinFixture
from lemoncheesecake.metadatapolicy import MetadataPolicy
from lemoncheesecake.reporting import get_reporting_backends, ReportingBackend
from lemoncheesecake.reporting.reportdir import create_report_dir_with_rotation
from lemoncheesecake.exceptions import ProjectLoadingError, ProjectNotFound, ModuleImportError
from lemoncheesecake.helpers.resources import get_resource_path
from lemoncheesecake.helpers.moduleimport import import_module
from lemoncheesecake.testtree import flatten_tests
from lemoncheesecake.exceptions import UserError, LemoncheesecakeException, serialize_current_exception

PROJECT_FILE = "project.py"
DEFAULT_REPORTING_BACKENDS = ("console", "json", "html")
DEFAULT_SUITES_DIR = "suites"
DEFAULT_FIXTURES_DIR = "fixtures"


def get_caller_dir(stack):
    frame_info = stack[1]
    return os.path.dirname(frame_info[1])


class Project:
    def __init__(self, project_dir=None) -> None:
        #: The project's directory path (optional, defaults to the caller dir)
        self.dir: str = osp.abspath(project_dir or get_caller_dir(inspect.stack()))
        #: The project's metadata policy
        self.metadata_policy: MetadataPolicy = MetadataPolicy()
        #: Indicates whether or not the project supports parallel execution of tests
        self.threaded: bool = True
        #: Indicated whether or not the command line ("lcc run...") will be displayed in the report
        self.show_command_line_in_report: bool = True
        #: The reporting backends of the project as a dict (whose key is the reporting backend name)
        self.reporting_backends: Dict[str, ReportingBackend] = {b.get_name(): b for b in get_reporting_backends()}
        #: The list of default reporting backend (indicated by their name) that will be used by "lcc run"
        self.default_reporting_backend_names = list(DEFAULT_REPORTING_BACKENDS)

    def add_cli_args(self, cli_parser: argparse.ArgumentParser) -> None:
        """
        Overridable. This method can be used to add extra CLI arguments to "lcc run".
        """
        pass

    def create_report_dir(self) -> str:
        """
        Overridable. Create the report directory when no report directory is specified to "lcc run".
        """
        return create_report_dir_with_rotation(self.dir)

    def load_suites(self) -> Sequence[Suite]:
        """
        Overridable. Load the project's suites.
        """
        return load_suites_from_directory(osp.join(self.dir, DEFAULT_SUITES_DIR))

    def load_fixtures(self) -> Sequence[Fixture]:
        """
        Overridable. Load the project's fixtures.
        """
        fixtures_dir = osp.join(self.dir, DEFAULT_FIXTURES_DIR)
        if osp.exists(fixtures_dir):
            return load_fixtures_from_directory(fixtures_dir)
        else:
            return []

    def pre_run(self, cli_args: Any, report_dir: str) -> None:
        """
        Overridable. This hook is called before running the tests.
        """
        pass

    def post_run(self, cli_args: Any, report_dir: str) -> None:
        """
        Overridable. This hook is called after running the tests.
        """
        pass

    def build_report_title(self) -> Optional[str]:
        """
        Overridable. Build a custom report title as a string.
        """
        return None

    def build_report_info(self) -> List[Tuple[str, str]]:
        """
        Overridable. Build a list key/value pairs (expressed as a two items tuple)
        that will be available in the report.

        Example::

            [
                ("key1", "value1"),
                ("key2", "value2")
            ]
        """
        info = []
        if self.show_command_line_in_report:
            info.append(("Command line", " ".join([os.path.basename(sys.argv[0])] + sys.argv[1:])))
        return info


def create_project(project_dir: str) -> None:
    shutil.copyfile(get_resource_path(("project", "template.py")), osp.join(project_dir, PROJECT_FILE))
    os.mkdir(osp.join(project_dir, "suites"))
    os.mkdir(osp.join(project_dir, "fixtures"))


def _load_project_from_file(path: str) -> Project:
    # load project module
    try:
        project_module = import_module(path)
    except ModuleImportError as e:
        raise ProjectLoadingError(str(e))

    # get project config instance
    try:
        project = project_module.project
    except AttributeError:
        raise ProjectLoadingError("Cannot find symbol 'project' in module '%s'" % path)
    if not isinstance(project, Project):
        raise ProjectLoadingError("Symbol 'project' in module '%s' does not inherit lemoncheesecake.project.Project" % path)
    
    return project


def _iter_on_path_hierarchy(path):
    yield path
    parent_path = osp.dirname(path)
    while parent_path != path:
        yield parent_path
        path, parent_path = parent_path, osp.dirname(parent_path)


def _load_project_from_path(path):
    if osp.isfile(path):
        return _load_project_from_file(path)
    elif osp.isdir(path):
        if osp.exists(osp.join(path, PROJECT_FILE)):
            return _load_project_from_file(osp.join(path, PROJECT_FILE))
        if osp.exists(osp.join(path, DEFAULT_SUITES_DIR)):
            return Project(path)

    raise ProjectLoadingError("'%s' is not a suitable project path" % path)


def load_project(path: str = None) -> Project:
    # first try: from argument
    if path:
        return _load_project_from_path(path)

    # second try: from environment
    path = os.environ.get("LCC_PROJECT", os.environ.get("LCC_PROJECT_FILE"))
    if path:
        return _load_project_from_path(path)

    # third try: look for a project.py file in directory hierarchy
    for dirname in _iter_on_path_hierarchy(os.getcwd()):
        filename = osp.join(dirname, PROJECT_FILE)
        if osp.exists(filename):
            return _load_project_from_file(filename)

    # fourth try: look for a "suites" sub-directory in the directory hierarchy
    for dirname in _iter_on_path_hierarchy(os.getcwd()):
        if osp.exists(osp.join(dirname, DEFAULT_SUITES_DIR)):
            return Project(dirname)

    # no luck
    raise ProjectNotFound("Cannot neither find a 'suites' directory nor a 'project.py' file")


class PreparedProject:
    def __init__(self, project, suites: Sequence[Suite], fixture_registry: FixtureRegistry, cli_args):
        self.project = project
        self.suites = suites
        self.fixture_registry = fixture_registry
        self.cli_args = cli_args

    @staticmethod
    def _build_fixture_registry(project, cli_args):
        registry = FixtureRegistry()
        registry.add_fixture(BuiltinFixture("cli_args", cli_args))
        registry.add_fixture(BuiltinFixture("project_dir", project.dir))
        for fixture in project.load_fixtures():
            registry.add_fixture(fixture)
        return registry

    @classmethod
    def create(cls, project: Project, suites: Sequence[Suite] = None,  cli_args=()):
        all_suites = project.load_suites()
        if suites is None:
            suites = all_suites

        project.metadata_policy.check_suites_compliance(suites)
        resolve_tests_dependencies(suites, all_suites)
        fixture_registry = cls._build_fixture_registry(project, cli_args)
        fixture_registry.check_dependencies()
        fixture_registry.check_fixtures_in_suites(suites)
        return cls(project, suites, fixture_registry, cli_args)

    def _setup_report(self, report):
        try:
            title = self.project.build_report_title()
        except Exception:
            raise LemoncheesecakeException(
                "Got an unexpected exception while getting report title from project:%s" % \
                serialize_current_exception(show_stacktrace=True)
            )
        if title:
            report.title = title

        try:
            info = list(self.project.build_report_info())
        except Exception:
            raise LemoncheesecakeException(
                "Got an unexpected exception while getting report info from project:%s" % \
                serialize_current_exception(show_stacktrace=True)
            )

        for key, value in info:
            report.add_info(key, value)

    def run(self, reporting_backends, report_dir, report_saving_strategy,
            force_disabled=False, stop_on_failure=False, nb_threads=1):
        # Handle "pre_run" hook
        try:
            self.project.pre_run(self.cli_args, report_dir)
        except UserError as e:
            raise e
        except Exception:
            raise LemoncheesecakeException(
                "Got an unexpected exception while running project's pre_run method:%s" % \
                    serialize_current_exception(show_stacktrace=True)
            )

        # Create session
        session = Session.create(
            AsyncEventManager.load(), reporting_backends, report_dir, report_saving_strategy,
            nb_threads=nb_threads, parallelized=nb_threads > 1 and len(list(flatten_tests(self.suites))) > 1
        )
        self._setup_report(session.report)

        # Run tests
        run_suites(
            self.suites, self.fixture_registry, session,
            force_disabled=force_disabled, stop_on_failure=stop_on_failure,
            nb_threads=nb_threads
        )

        # Handle "post_run" hook
        try:
            self.project.post_run(self.cli_args, report_dir)
        except UserError as e:
            raise e
        except Exception:
            raise LemoncheesecakeException(
                "Got an unexpected exception while running project's post_run method:%s" % \
                    serialize_current_exception(show_stacktrace=True)
            )

        return session.report



'''
Created on Sep 8, 2016

@author: nicolas
'''

import re
import fnmatch
from functools import reduce

from lemoncheesecake.reporting import load_report
from lemoncheesecake.reporting.reportdir import DEFAULT_REPORT_DIR_NAME
from lemoncheesecake.reporting.report import Result, TestResult, Step, Log, Check, Attachment, Url
from lemoncheesecake.testtree import BaseTest, BaseSuite
from lemoncheesecake.suite import Test
from lemoncheesecake.exceptions import UserError

_NEGATION_FLAGS = "-^~"


def _iter_grepable(steps):
    for step in steps:
        yield step.description
        for log in step.get_logs():
            if isinstance(log, Log):
                yield log.message
            elif isinstance(log, Check):
                yield log.description
                if log.details:
                    yield log.details
            elif isinstance(log, Attachment):
                yield log.filename
                yield log.description
            elif isinstance(log, Url):
                yield log.url
                yield log.description


def _grep(pattern, steps):
    return any(map(pattern.search, _iter_grepable(steps)))


class Filter:
    def __bool__(self):
        raise NotImplementedError()

    def __call__(self, test):
        raise NotImplementedError()


class BaseTreeNodeFilter(Filter):
    def __init__(self, paths=(), descriptions=(), tags=(), properties=(), links=()):
        self.paths = list(paths)
        self.descriptions = list(descriptions)
        self.tags = list(tags)
        self.properties = list(properties)
        self.links = list(links)

    def __bool__(self):
        return any((
            self.paths, self.descriptions, self.tags, self.properties, self.links
        ))

    @staticmethod
    def _match_values(values, patterns):
        if not patterns:
            return True

        values = [value or "" for value in values]  # convert None to ""

        for pattern in patterns:
            if pattern[0] in _NEGATION_FLAGS:
                if not fnmatch.filter(values, pattern[1:]):
                    return True
            else:
                if fnmatch.filter(values, pattern):
                    return True
        return False

    @staticmethod
    def _match_key_values(key_values, patterns):
        if not patterns:
            return True

        for key, value in patterns:
            if key in key_values:
                if value[0] in _NEGATION_FLAGS:
                    if not fnmatch.fnmatch(key_values[key], value[1:]):
                        return True
                else:
                    if fnmatch.fnmatch(key_values[key], value):
                        return True
        return False

    @staticmethod
    def _match_values_lists(lsts, patterns):
        return BaseTreeNodeFilter._match_values(
            reduce(lambda x, y: list(x) + list(y), lsts, []),  # make a flat list
            patterns
        )

    def _do_paths(self, node):
        return self._match_values(node.hierarchy_paths, self.paths)

    def _do_descriptions(self, node):
        return all(self._match_values(node.hierarchy_descriptions, descs) for descs in self.descriptions)

    def _do_tags(self, node):
        return all(self._match_values(node.hierarchy_tags, tags) for tags in self.tags)

    def _do_properties(self, node):
        return all(self._match_key_values(node.hierarchy_properties, props) for props in self.properties)

    def _do_links(self, node):
        return all(self._match_values_lists(node.hierarchy_links, links) for links in self.links)

    @staticmethod
    def _apply_criteria(obj, *criteria):
        return all(criterion(obj) for criterion in criteria)

    def __call__(self, node):
        assert isinstance(node, (BaseTest, BaseSuite))
        return self._apply_criteria(
            node, self._do_paths, self._do_descriptions, self._do_tags, self._do_properties, self._do_links
        )


class TestFilter(BaseTreeNodeFilter):
    def __init__(self, enabled=False, disabled=False, **kwargs):
        BaseTreeNodeFilter.__init__(self, **kwargs)
        self.enabled = enabled
        self.disabled = disabled

    def __bool__(self):
        return BaseTreeNodeFilter.__bool__(self) or any((self.enabled, self.disabled))

    def _do_enabled(self, test):
        return not test.is_disabled() if self.enabled else True

    def _do_disabled(self, test):
        return test.is_disabled() if self.disabled else True

    def _apply_test_criteria(self, test):
        return self._apply_criteria(test, self._do_enabled, self._do_disabled)

    def __call__(self, test):
        assert isinstance(test, Test)
        return BaseTreeNodeFilter.__call__(self, test) and self._apply_test_criteria(test)


class ResultFilter(BaseTreeNodeFilter):
    def __init__(self, statuses=None, enabled=False, disabled=False, grep=None, **kwargs):
        BaseTreeNodeFilter.__init__(self, **kwargs)
        self.statuses = set(statuses) if statuses is not None else set()
        self.enabled = enabled
        self.disabled = disabled
        self.grep = grep

    def __bool__(self):
        return BaseTreeNodeFilter.__bool__(self) or any((self.statuses, self.enabled, self.disabled, self.grep))

    def _do_statuses(self, result):
        return result.status in self.statuses if self.statuses else True

    def _do_enabled(self, result):
        return result.status != "disabled" if self.enabled else True

    def _do_disabled(self, result):
        return result.status == "disabled" if self.disabled else True

    def _do_grep(self, result):
        if not self.grep:
            return True

        return _grep(self.grep, result.get_steps())

    def _apply_result_criteria(self, result):
        return self._apply_criteria(
            result, self._do_statuses, self._do_enabled, self._do_disabled, self._do_grep
        )

    def __call__(self, result: Result) -> bool:
        assert isinstance(result, Result)

        # test result:
        if isinstance(result, TestResult):
            return BaseTreeNodeFilter.__call__(self, result) and self._apply_result_criteria(result)
        # suite setup or teardown result, apply the base filter on the suite node:
        elif result.parent_suite:
            return BaseTreeNodeFilter.__call__(self, result.parent_suite) and self._apply_result_criteria(result)
        # session setup or teardown:
        else:
            if BaseTreeNodeFilter.__bool__(self):
                # no criteria of BaseFilter is applicable to a session setup/teardown result,
                # meaning it's a no match
                return False
            else:
                return self._apply_result_criteria(result)


class StepFilter(BaseTreeNodeFilter):
    def __init__(self, passed=False, failed=False, grep=None, **kwargs):
        BaseTreeNodeFilter.__init__(self, **kwargs)
        self.passed = passed
        self.failed = failed
        self.grep = grep

    def __bool__(self):
        return BaseTreeNodeFilter.__bool__(self) or any((self.passed, self.failed, self.grep))

    def _do_passed(self, step):
        return step.is_successful() if self.passed else True

    def _do_failed(self, step):
        return not step.is_successful() if self.failed else True

    def _do_grep(self, step):
        if not self.grep:
            return True

        return _grep(self.grep, (step,))

    def _apply_step_criteria(self, step):
        return self._apply_criteria(
            step, self._do_passed, self._do_failed, self._do_grep
        )

    def __call__(self, step: Step) -> bool:
        assert isinstance(step, Step)

        # test result:
        if isinstance(step.parent_result, TestResult):
            return BaseTreeNodeFilter.__call__(self, step.parent_result) and self._apply_step_criteria(step)
        # suite setup or teardown result, apply the base filter on the suite node:
        elif step.parent_result.parent_suite:
            return BaseTreeNodeFilter.__call__(self, step.parent_result.parent_suite) and self._apply_step_criteria(step)
        # session setup or teardown:
        else:
            if BaseTreeNodeFilter.__bool__(self):
                # no criteria of BaseFilter is applicable to a session setup/teardown result,
                # meaning it's a no match
                return False
            else:
                return self._apply_step_criteria(step)


class FromTestsFilter(Filter):
    def __init__(self, tests):
        self._tests = [test.path for test in tests]

    def __bool__(self):
        return True

    def __call__(self, test):
        return test.path in self._tests


def _add_filter_cli_args(cli_parser, no_positional_argument=False, only_executed_tests=False):
    def property_value(value):
        splitted = value.split(":")
        if len(splitted) != 2:
            raise ValueError()
        return splitted

    group = cli_parser.add_argument_group("Filtering")
    if no_positional_argument:
        group.add_argument(
            "--path", "-p", nargs="+", help="Filter on test/suite path (wildcard character '*' can be used)"
        )
    else:
        group.add_argument(
            "path", nargs="*", default=[], help="Filter on test/suite path (wildcard character '*' can be used)"
        )
    group.add_argument(
        "--desc", nargs="+", action="append", default=[], help="Filter on descriptions"
    )
    group.add_argument(
        "--tag", "-a", nargs="+", action="append", default=[], help="Filter on tags"
    )
    group.add_argument(
        "--property", "-m", nargs="+", type=property_value, action="append", default=[], help="Filter on properties"
    )
    group.add_argument(
        "--link", "-l", nargs="+", action="append", default=[], help="Filter on links (names and URLs)"
    )
    group.add_argument(
        "--passed", action="store_true", help="Filter on passed tests"
    )
    group.add_argument(
        "--failed", action="store_true", help="Filter on failed tests"
    )
    group.add_argument(
        "--grep", "-g", help="Filter result content using pattern"
    )
    if not only_executed_tests:
        group.add_argument(
            "--skipped", action="store_true", help="Filter on skipped tests"
        )
        group.add_argument(
            "--non-passed", action="store_true", help="Alias for --failed --skipped"
        )
        group.add_argument(
            "--disabled", action="store_true", help="Filter on disabled tests"
        )
        group.add_argument(
            "--enabled", action="store_true", help="Filter on enabled (non-disabled) tests"
        )

    return group


def add_test_filter_cli_args(cli_parser):
    group = _add_filter_cli_args(cli_parser)
    group.add_argument(
        "--from-report", required=False, help="When enabled, the filtering is based on the given report"
    )
    return group


def add_result_filter_cli_args(cli_parser, only_executed_tests=False):
    return _add_filter_cli_args(cli_parser, no_positional_argument=True, only_executed_tests=only_executed_tests)


def add_step_filter_cli_args(cli_parser):
    return _add_filter_cli_args(cli_parser, no_positional_argument=True, only_executed_tests=True)


def _set_common_filter_criteria(fltr, cli_args, only_executed_tests=False):
    if not only_executed_tests and (cli_args.disabled and cli_args.enabled):
        raise UserError("--disabled and --enabled arguments are mutually exclusive")

    fltr.paths = cli_args.path
    fltr.descriptions = cli_args.desc
    fltr.tags = cli_args.tag
    fltr.properties = cli_args.property
    fltr.links = cli_args.link
    if not only_executed_tests:
        fltr.disabled = cli_args.disabled
        fltr.enabled = cli_args.enabled


def _make_test_filter(cli_args):
    if cli_args.passed or cli_args.failed or cli_args.skipped:
        raise UserError("--passed, --failed and --skipped arguments can only be used on the report-based filter")
    test_filter = TestFilter()
    _set_common_filter_criteria(test_filter, cli_args)
    return test_filter


def _make_grep_criterion(grep):
    return re.compile(grep, re.IGNORECASE | re.MULTILINE)


def make_result_filter(cli_args, only_executed_tests=False):
    result_filter = ResultFilter()
    _set_common_filter_criteria(result_filter, cli_args, only_executed_tests=only_executed_tests)

    if only_executed_tests:
        if cli_args.passed:
            result_filter.statuses.add("passed")
        if cli_args.failed:
            result_filter.statuses.add("failed")
        # when neither --passed not --failed was passed, enforce statuses passed and failed
        # to select tests that have been executed
        if not result_filter.statuses:
            result_filter.statuses.update(("passed", "failed"))
    else:
        if cli_args.passed:
            result_filter.statuses.add("passed")
        if cli_args.failed:
            result_filter.statuses.add("failed")
        if cli_args.skipped:
            result_filter.statuses.add("skipped")
        if cli_args.non_passed:
            result_filter.statuses.update(("failed", "skipped"))

    if cli_args.grep:
        result_filter.grep = _make_grep_criterion(cli_args.grep)

    return result_filter


def _make_from_report_filter(cli_args, only_executed_tests=False):
    report = load_report(cli_args.from_report or DEFAULT_REPORT_DIR_NAME)
    test_filter = make_result_filter(cli_args, only_executed_tests=only_executed_tests)
    return FromTestsFilter(filter(test_filter, report.all_tests()))


def make_test_filter(cli_args):
    if any((cli_args.from_report, cli_args.passed, cli_args.failed, cli_args.skipped, cli_args.non_passed, cli_args.grep)):
        return _make_from_report_filter(cli_args)
    else:
        return _make_test_filter(cli_args)


def make_step_filter(cli_args):
    if cli_args.passed and cli_args.failed:
        raise UserError("--passed and --failed arguments are mutually exclusive")

    step_filter = StepFilter()
    _set_common_filter_criteria(step_filter, cli_args, only_executed_tests=True)

    step_filter.passed = cli_args.passed
    step_filter.failed = cli_args.failed
    if cli_args.grep:
        step_filter.grep = _make_grep_criterion(cli_args.grep)

    return step_filter



from __future__ import annotations

import inspect
from typing import List, Any, Sequence, Callable, Optional

from lemoncheesecake.helpers.moduleimport import import_module, get_matching_files, get_py_files_from_dir
from lemoncheesecake.exceptions import ValidationError, ModuleImportError, FixtureLoadingError
from lemoncheesecake.helpers.orderedset import OrderedSet
from lemoncheesecake.helpers.introspection import get_callable_args
from lemoncheesecake.helpers.threading import ThreadedFactory


_FORBIDDEN_FIXTURE_NAMES = ("fixture_name",)
_SCOPE_LEVELS = {
    "test": 1,
    "suite": 2,
    "session": 3,
    "pre_run": 4
}

_scheduled_fixtures: Optional[ScheduledFixtures] = None


def initialize_fixture_cache(scheduled_fixtures):
    global _scheduled_fixtures
    _scheduled_fixtures = scheduled_fixtures


def get_fixture(name):
    """
    Return the corresponding fixture value. Only fixtures whose scope is 'pre_run' can be retrieved.
    """
    global _scheduled_fixtures

    assert _scheduled_fixtures, "Fixture cache has not yet been initialized"
    if not _scheduled_fixtures.has_fixture(name):
        raise LookupError("Fixture '%s' either does not exist or doesn't have a pre_run scope" % name)

    return _scheduled_fixtures.get_fixture_result(name)


class _FixtureInfo:
    def __init__(self, names, scope, per_thread):
        self.names = names
        self.scope = scope
        self.per_thread = per_thread


def fixture(names=None, scope="test", per_thread=False):
    """
    Decorator, declare a function as a fixture.

    :param names: an optional list of names that can be used to access the fixture value,
        if no names are provided the decorated function name will be used

    :param scope: the fixture scope, available scopes are: ``test``, ``suite``, ``session``, ``pre_run``;
        default is ``test``

    :param per_thread: whether or not the fixture must be executed on a per-thread basis
        Please note that when ``per_thread`` is set to ``True``:

        - the scope can only be ``session`` or ``suite``
        - the fixture can only be used in tests or by fixtures with the ``test`` scope
    """
    def wrapper(func):
        if scope not in _SCOPE_LEVELS.keys():
            raise ValueError("Invalid fixture scope '%s' in fixture function '%s'" % (scope, func.__name__))

        if per_thread and scope not in ("session", "suite"):
            raise AssertionError("The fixture can only be per_thread=True if scope is 'session' or 'suite'")

        setattr(func, "_lccfixtureinfo", _FixtureInfo(names or [func.__name__], scope, per_thread))
        return func

    return wrapper


class _BaseFixtureResult:
    def get(self):
        raise NotImplementedError()

    def teardown(self):
        pass


class _FixtureResult(_BaseFixtureResult):
    def __init__(self, value):
        self.value = value

    def get(self):
        return self.value


class _GeneratorFixtureResult(_BaseFixtureResult):
    def __init__(self, name, generator):
        self.name = name
        self.generator = generator
        self.value = next(generator)

    def get(self):
        return self.value

    def teardown(self):
        try:
            next(self.generator)
        except StopIteration:
            pass
        else:
            raise AssertionError(
                "Fixture '%s' yields more than once: only one yield is supported." % self.name
            )


class _PerThreadFixtureResult(_BaseFixtureResult, ThreadedFactory):
    def __init__(self, name, func, params):
        ThreadedFactory.__init__(self)
        self.name = name
        self.func = func
        self.params = params

    def get(self):
        return self.get_object().get()

    def teardown(self):
        self.teardown_factory()

    def setup_object(self):
        return _build_fixture_result_from_func(self.name, self.func, self.params)

    def teardown_object(self, result):
        result.teardown()


def _build_fixture_result_from_func(name, func, params, per_thread=False):
    if per_thread:
        return _PerThreadFixtureResult(name, func, params)
    else:
        value = func(**params)
        if inspect.isgenerator(value):
            return _GeneratorFixtureResult(name, value)
        else:
            return _FixtureResult(value)


class _BaseFixture:
    def __init__(self, name, scope, params, per_thread):
        self.name = name
        self.scope = scope
        self.params = params
        self.per_thread = per_thread

    @property
    def scope_level(self):
        return _SCOPE_LEVELS[self.scope]

    def execute(self, params: dict) -> _BaseFixtureResult:
        raise NotImplementedError()


class Fixture(_BaseFixture):
    def __init__(self, name, func, scope, params, per_thread):
        _BaseFixture.__init__(self, name, scope, params, per_thread)
        self.func = func

    def execute(self, params):
        for param_name in params.keys():
            assert param_name in self.params
        return _build_fixture_result_from_func(self.name, self.func, params, self.per_thread)


class BuiltinFixture(_BaseFixture):
    def __init__(self, name, value):
        _BaseFixture.__init__(self, name, scope="pre_run", params={}, per_thread=False)
        self.result = _FixtureResult(value)

    def execute(self, _):
        return self.result


class ScheduledFixtures:
    def __init__(self, scope, fixtures=(), parent_scheduled_fixtures=None):
        self.scope = scope
        self._fixtures = {}
        self._parent_scheduled_fixtures = parent_scheduled_fixtures
        self._results = {}
        for fixture in fixtures:
            self.add_fixture(fixture)

    def add_fixture(self, fixture):
        assert fixture.scope == self.scope
        self._fixtures[fixture.name] = fixture

    def get_fixture_names(self):
        return self._fixtures.keys()

    def has_fixture(self, name):
        return name in self._fixtures

    def is_empty(self):
        return len(self._fixtures) == 0

    def _get_fixture_params(self, name):
        return {
            param_name: name if param_name == "fixture_name" else self.get_fixture_result(param_name)
                for param_name in self._fixtures[name].params
        }

    def _setup_fixture(self, name):
        assert name not in self._results, "Cannot setup fixture '%s', it has already been executed" % name
        self._results[name] = self._fixtures[name].execute(self._get_fixture_params(name))

    def _teardown_fixture(self, name):
        assert name in self._results, "Cannot teardown fixture '%s', it has not been previously executed" % name
        self._results[name].teardown()
        del self._results[name]

    def get_setup_teardown_pairs(self):
        return list(map(
            lambda name: (lambda: self._setup_fixture(name), lambda: self._teardown_fixture(name)),
            self._fixtures
        ))

    def get_fixture_result(self, name):
        if name in self._fixtures:
            assert name in self._results, "Cannot get fixture '%s' result, it has not been previously executed" % name
            return self._results[name].get()
        elif self._parent_scheduled_fixtures:
            return self._parent_scheduled_fixtures.get_fixture_result(name)
        else:
            raise LookupError("Cannot find fixture named '%s' in scheduled fixtures" % name)

    def get_fixture_results(self, names):
        return {name: self.get_fixture_result(name) for name in names}


class FixtureRegistry:
    def __init__(self):
        self._fixtures = {}

    def add_fixture(self, fixture):
        if fixture.name in self._fixtures and isinstance(self._fixtures[fixture.name], BuiltinFixture):
            raise ValidationError("'%s' is a builtin fixture name" % fixture.name)
        self._fixtures[fixture.name] = fixture

    def add_fixtures(self, fixtures):
        for fixture in fixtures:
            self.add_fixture(fixture)

    def get_fixture(self, name):
        return self._fixtures[name]

    def get_fixture_dependencies(self, name, ref_fixtures=()):
        fixture_params = [p for p in self._fixtures[name].params if p != "fixture_name"]
        if any(ref_fixture in fixture_params for ref_fixture in ref_fixtures):
            raise ValidationError(
                "Fixture params %s have circular dependency on a fixture among %s" % (fixture_params, ref_fixtures)
            )

        dependencies = OrderedSet()
        for param in fixture_params:
            if param not in self._fixtures:
                raise ValidationError("Fixture '%s' used by fixture '%s' does not exist" % (param, name))
            dependencies.update(self.get_fixture_dependencies(param, (name,) + ref_fixtures))
        dependencies.update(fixture_params)

        return dependencies

    def check_dependencies(self):
        """
        Checks for:
        - missing dependencies
        - circular dependencies
        - scope incoherence
        - forbidden fixture name
        raises FixtureConstraintViolation if a check fails
        """

        # first, check for forbidden fixture name
        for fixture_name in self._fixtures.keys():
            if fixture_name in _FORBIDDEN_FIXTURE_NAMES:
                raise ValidationError("Fixture name '%s' is forbidden" % fixture_name)

        # second, check for missing & circular dependencies
        for fixture_name in self._fixtures.keys():
            self.get_fixture_dependencies(fixture_name)

        # third, check fixture compliance with their direct fixture dependencies
        for fixture in self._fixtures.values():
            dependency_fixtures = [self._fixtures[param] for param in fixture.params if param != "fixture_name"]
            for dependency_fixture in dependency_fixtures:
                if dependency_fixture.per_thread and fixture.scope != "test":
                    raise ValidationError(
                        "Fixture '%s' with scope '%s' is incompatible with per-thread fixture '%s'" % (
                            fixture.name, fixture.scope, dependency_fixture.name
                        )
                    )
                if dependency_fixture.scope_level < fixture.scope_level:
                    raise ValidationError(
                        "Fixture '%s' with scope '%s' is incompatible with scope '%s' of fixture '%s'" % (
                            fixture.name, fixture.scope, dependency_fixture.scope, dependency_fixture.name
                        )
                    )

    def check_fixtures_in_test(self, test):
        for fixture in test.get_fixtures():
            if fixture not in self._fixtures:
                raise ValidationError("Unknown fixture '%s' used in test '%s'" % (fixture, test.path))

    def check_fixtures_in_suite(self, suite):
        for fixture_name in suite.get_fixtures():
            try:
                fixture = self._fixtures[fixture_name]
            except KeyError:
                raise ValidationError("Suite '%s' uses an unknown fixture '%s'" % (suite.path, fixture_name))
            if fixture.per_thread:
                raise ValidationError(
                    "Suite '%s' uses per-thread fixture '%s' which is not allowed" % (
                        suite.path, fixture.name
                    )
                )
            if fixture.scope_level < _SCOPE_LEVELS["suite"]:
                raise ValidationError("Suite '%s' uses fixture '%s' which has an incompatible scope" % (
                    suite.path, fixture.name
                ))

        for test in suite.get_tests():
            self.check_fixtures_in_test(test)

        for sub_suite in suite.get_suites():
            self.check_fixtures_in_suite(sub_suite)

    def check_fixtures_in_suites(self, suites):
        for suite in suites:
            self.check_fixtures_in_suite(suite)

    def get_fixture_scope(self, name):
        return self._fixtures[name].scope

    @staticmethod
    def get_fixtures_used_in_suite(suite, include_disabled):
        if not suite.has_enabled_tests() and not include_disabled:
            return OrderedSet()

        fixtures = suite.get_fixtures()

        for test in suite.get_tests():
            if test.is_enabled() or include_disabled:
                fixtures.update(test.get_fixtures())

        return fixtures

    @staticmethod
    def get_fixtures_used_in_suite_recursively(suite, include_disabled):
        fixtures = FixtureRegistry.get_fixtures_used_in_suite(suite, include_disabled)

        for sub_suite in suite.get_suites():
            fixtures.update(FixtureRegistry.get_fixtures_used_in_suite_recursively(sub_suite, include_disabled))

        return fixtures

    def get_scheduled_fixtures_for_scope(self, direct_fixtures, scope, parent_scheduled_fixtures=None):
        fixtures = OrderedSet()

        for fixture in direct_fixtures:
            fixtures.update(self.get_fixture_dependencies(fixture))
            fixtures.add(fixture)

        return ScheduledFixtures(
            scope, [self._fixtures[name] for name in fixtures if self._fixtures[name].scope == scope],
            parent_scheduled_fixtures=parent_scheduled_fixtures
        )

    def get_fixtures_scheduled_for_pre_run(self, suites, include_disabled=False):
        fixtures = OrderedSet()
        for suite in suites:
            fixtures.update(FixtureRegistry.get_fixtures_used_in_suite_recursively(suite, include_disabled))
        return self.get_scheduled_fixtures_for_scope(fixtures, "pre_run")

    def get_fixtures_scheduled_for_session(self, suites, prerun_session_scheduled_fixtures, include_disabled=False):
        fixtures = OrderedSet()
        for suite in suites:
            fixtures.update(FixtureRegistry.get_fixtures_used_in_suite_recursively(suite, include_disabled))
        return self.get_scheduled_fixtures_for_scope(fixtures, "session", prerun_session_scheduled_fixtures)

    def get_fixtures_scheduled_for_suite(self, suite, session_scheduled_fixtures, include_disabled=False):
        return self.get_scheduled_fixtures_for_scope(
            FixtureRegistry.get_fixtures_used_in_suite(suite, include_disabled), "suite", session_scheduled_fixtures
        )

    def get_fixtures_scheduled_for_test(self, test, suite_scheduled_fixtures):
        return self.get_scheduled_fixtures_for_scope(
            test.get_fixtures(), "test", suite_scheduled_fixtures
        )


def load_fixtures_from_func(func: Callable) -> List[Fixture]:
    """
    Load a fixture from a function that has been decorated with ``@lcc.fixture()``
    """
    assert hasattr(func, "_lccfixtureinfo")
    info = func._lccfixtureinfo  # noqa
    args = get_callable_args(func)
    return [Fixture(name, func, info.scope, args, info.per_thread) for name in info.names]


def load_fixtures_from_module(mod: Any) -> List[Fixture]:
    """
    Load fixtures from a module instance.

    .. versionadded:: 1.5.1
    """
    fixtures = []
    for sym_name in dir(mod):
        sym = getattr(mod, sym_name)
        if hasattr(sym, "_lccfixtureinfo"):
            fixtures.extend(load_fixtures_from_func(sym))

    return fixtures


def load_fixtures_from_file(filename: str) -> List[Fixture]:
    """
    Load fixtures from a given file.
    """
    try:
        mod = import_module(filename)
    except ModuleImportError as e:
        raise FixtureLoadingError(str(e))

    return load_fixtures_from_module(mod)


def load_fixtures_from_files(patterns: Any[str, Sequence[str]], excluding: Any[str, Sequence[str]] = []) -> List[Fixture]:
    """
    Load fixtures from files.

    :param patterns: a mandatory list (a simple string can also be used instead of a single element list)
        of files to import; the wildcard '*' character can be used
    :param excluding: an optional list (a simple string can also be used instead of a single element list)
      of elements to exclude from the expanded list of files to import

    Example::

        load_fixtures_from_files("test_*.py")
    """
    fixtures = []
    for file in get_matching_files(patterns, excluding):
        fixtures.extend(load_fixtures_from_file(file))
    return fixtures


def load_fixtures_from_directory(dir: str) -> List[Fixture]:
    """
    Load fixtures from a given directory (not recursive).
    """
    fixtures = []
    for file in get_py_files_from_dir(dir):
        fixtures.extend(load_fixtures_from_file(file))
    return fixtures
